from .Renderer import Renderer
from ..PostProcessEffects import FXAAEffect
from ..Frame import Frame
from ..GlassEngineConfig import GlassEngineConfig

from glass import ShaderProgram, GLConfig, FBO, sampler2D, sampler2DMS, samplerCube, sampler2DArray, GlassConfig, GLInfo
from glass.utils import cat, modify_time

from OpenGL import GL
import glm
import os

class CommonRenderer(Renderer):

    __dir_light_depth_geo_shader_template = None
    __dir_light_depth_lines_geo_shader_template = None
    __dir_light_depth_points_geo_shader_template = None

    def __init__(self):
        Renderer.__init__(self)

        self._depth_map = None
        self._view_pos_map = None
        self._view_normal_map = None

        self._should_update = False
        self._all_meshes = []
        self._all_lines = []
        self._all_points = []
        self._opaque_meshes = []
        self._opaque_lines = []
        self._opaque_points = []
        self._transparent_meshes = []
        self._transparent_points = []
        self._transparent_lines = []
        self._meshes_cast_shadows = []
        self._lines_cast_shadows = []
        self._points_cast_shadows = []

    @property
    def programs(self):
        if not hasattr(self, "_programs"):
            self._programs = {}
            
        return self._programs
    
    @property
    def fbos(self):
        if not hasattr(self, "_fbos"):
            self._fbos = {}

        return self._fbos
    
    @property
    def dir_light_depth_geo_shader_path(self):
        self_folder = os.path.dirname(os.path.abspath(__file__))
        target_filename = GlassConfig.cache_folder + f"/DirLight_depth{self.camera.CSM_levels}.gs"
        template_filename = self_folder + "/../glsl/Pipelines/DirLight_depth/DirLight_depth.gs"
        if not os.path.isfile(target_filename) or modify_time(template_filename) > modify_time(target_filename):
            if CommonRenderer.__dir_light_depth_geo_shader_template is None:
                CommonRenderer.__dir_light_depth_geo_shader_template = cat(template_filename)

            content = CommonRenderer.__dir_light_depth_geo_shader_template.replace("{CSM_levels}", str(self.camera.CSM_levels))
            out_file = open(target_filename, "w")
            out_file.write(content)
            out_file.close()

        return target_filename
    
    @property
    def dir_light_depth_lines_geo_shader_path(self):
        self_folder = os.path.dirname(os.path.abspath(__file__))
        target_filename = GlassConfig.cache_folder + f"/DirLight_depth_lines{self.camera.CSM_levels}.gs"
        template_filename = self_folder + "/../glsl/Pipelines/DirLight_depth/DirLight_depth_lines.gs"
        if not os.path.isfile(target_filename) or modify_time(template_filename) > modify_time(target_filename):
            if CommonRenderer.__dir_light_depth_lines_geo_shader_template is None:
                CommonRenderer.__dir_light_depth_lines_geo_shader_template = cat(template_filename)

            content = CommonRenderer.__dir_light_depth_lines_geo_shader_template.replace("{CSM_levels}", str(self.camera.CSM_levels))
            out_file = open(target_filename, "w")
            out_file.write(content)
            out_file.close()

        return target_filename

    @property
    def dir_light_depth_points_geo_shader_path(self):
        self_folder = os.path.dirname(os.path.abspath(__file__))
        target_filename = GlassConfig.cache_folder + f"/DirLight_depth_points{self.camera.CSM_levels}.gs"
        template_filename = self_folder + "/../glsl/Pipelines/DirLight_depth/DirLight_depth_points.gs"
        if not os.path.isfile(target_filename) or modify_time(template_filename) > modify_time(target_filename):
            if CommonRenderer.__dir_light_depth_points_geo_shader_template is None:
                CommonRenderer.__dir_light_depth_points_geo_shader_template = cat(template_filename)

            content = CommonRenderer.__dir_light_depth_points_geo_shader_template.replace("{CSM_levels}", str(self.camera.CSM_levels))
            out_file = open(target_filename, "w")
            out_file.write(content)
            out_file.close()

        return target_filename

    @property
    def dir_light_depth_program(self):
        key = f"dir_light_depth{self.camera.CSM_levels}"
        if key in self.programs:
            return self.programs[key]
        
        program = ShaderProgram()
        GlassEngineConfig.define_for_program(program)
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program.add_include_path(self_folder + "/../glsl/Pipelines/DirLight_depth")
        program.compile(self_folder + "/../glsl/Pipelines/DirLight_depth/DirLight_depth.vs")
        program.compile(self.dir_light_depth_geo_shader_path)
        program.compile(self_folder + "/../glsl/Pipelines/DirLight_depth/DirLight_depth.fs")

        self.programs[key] = program

        return program
    
    @property
    def dir_light_depth_lines_program(self):
        key = f"dir_light_depth_lines{self.camera.CSM_levels}"
        if key in self.programs:
            return self.programs[key]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        GlassEngineConfig.define_for_program(program)
        program.add_include_path(self_folder + "/../glsl/Pipelines/DirLight_depth")
        program.compile(self_folder + "/../glsl/Pipelines/DirLight_depth/DirLight_depth.vs")
        program.compile(self.dir_light_depth_lines_geo_shader_path)
        program.compile(self_folder + "/../glsl/Pipelines/DirLight_depth/DirLight_depth.fs")

        self.programs[key] = program

        return program

    @property
    def dir_light_depth_points_program(self):
        key = f"dir_light_depth_points{self.camera.CSM_levels}"
        if key in self.programs:
            return self.programs[key]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        GlassEngineConfig.define_for_program(program)
        program.add_include_path(self_folder + "/../glsl/Pipelines/DirLight_depth")
        program.compile(self_folder + "/../glsl/Pipelines/DirLight_depth/DirLight_depth.vs")
        program.compile(self.dir_light_depth_points_geo_shader_path)
        program.compile(self_folder + "/../glsl/Pipelines/DirLight_depth/DirLight_depth.fs")

        self.programs[key] = program

        return program
    
    @property
    def point_light_depth_program(self):
        if "point_light_depth" in self.programs:
            return self.programs["point_light_depth"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        GlassEngineConfig.define_for_program(program)
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.vs")
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.gs")
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.fs")

        self.programs["point_light_depth"] = program

        return program
    
    @property
    def point_light_depth_lines_program(self):
        if "point_light_depth_lines" in self.programs:
            return self.programs["point_light_depth_lines"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        GlassEngineConfig.define_for_program(program)
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.vs")
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth_lines.gs")
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.fs")

        self.programs["point_light_depth_lines"] = program

        return program

    @property
    def point_light_depth_points_program(self):
        if "point_light_depth_points" in self.programs:
            return self.programs["point_light_depth_points"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        GlassEngineConfig.define_for_program(program)
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.vs")
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth_points.gs")
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.fs")

        self.programs["point_light_depth_points"] = program

        return program
    
    @property
    def spot_light_depth_program(self):
        if "spot_light_depth" in self.programs:
            return self.programs["spot_light_depth"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        GlassEngineConfig.define_for_program(program)
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.vs")
        program.compile(self_folder + "/../glsl/Pipelines/SpotLight_depth/SpotLight_depth.gs")
        program.compile(self_folder + "/../glsl/Pipelines/SpotLight_depth/SpotLight_depth.fs")

        self.programs["spot_light_depth"] = program

        return program
    
    @property
    def spot_light_depth_lines_program(self):
        if "spot_light_depth_lines" in self.programs:
            return self.programs["spot_light_depth_lines"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        GlassEngineConfig.define_for_program(program)
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.vs")
        program.compile(self_folder + "/../glsl/Pipelines/SpotLight_depth/SpotLight_depth_lines.gs")
        program.compile(self_folder + "/../glsl/Pipelines/SpotLight_depth/SpotLight_depth.fs")

        self.programs["spot_light_depth_lines"] = program

        return program

    @property
    def spot_light_depth_points_program(self):
        if "spot_light_depth_points" in self.programs:
            return self.programs["spot_light_depth_points"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        GlassEngineConfig.define_for_program(program)
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.vs")
        program.compile(self_folder + "/../glsl/Pipelines/SpotLight_depth/SpotLight_depth_points.gs")
        program.compile(self_folder + "/../glsl/Pipelines/SpotLight_depth/SpotLight_depth.fs")

        self.programs["spot_light_depth_points"] = program

        return program
    
    def classify_meshes(self):
        self._all_meshes.clear()
        self._all_lines.clear()
        self._all_points.clear()
        self._opaque_meshes.clear()
        self._opaque_lines.clear()
        self._opaque_points.clear()
        self._transparent_meshes.clear()
        self._transparent_lines.clear()
        self._transparent_points.clear()
        self._meshes_cast_shadows.clear()
        self._lines_cast_shadows.clear()
        self._points_cast_shadows.clear()

        for mesh_tuple in self.scene.all_meshes.items():
            mesh = mesh_tuple[0]
            if mesh.primitive_type in GLInfo.triangle_types:
                self._all_meshes.append(mesh_tuple)
                
                if mesh.material.cast_shadows:
                    self._meshes_cast_shadows.append(mesh_tuple)

                if mesh.has_transparent:
                    self._transparent_meshes.append(mesh_tuple)

                if mesh.has_opaque:
                    self._opaque_meshes.append(mesh_tuple)

            elif mesh.primitive_type in GLInfo.line_types:
                self._all_lines.append(mesh_tuple)
                
                if mesh.material.cast_shadows:
                    self._lines_cast_shadows.append(mesh_tuple)

                if mesh.has_transparent:
                    self._transparent_lines.append(mesh_tuple)

                if mesh.has_opaque:
                    self._opaque_lines.append(mesh_tuple)

            elif mesh.primitive_type == GL.GL_POINTS:
                self._all_points.append(mesh_tuple)
                
                if mesh.material.cast_shadows:
                    self._points_cast_shadows.append(mesh_tuple)

                if mesh.has_transparent:
                    self._transparent_points.append(mesh_tuple)

                if mesh.has_opaque:
                    self._opaque_points.append(mesh_tuple)

    def update_spot_lights_depth(self):
        if "GL_ARB_bindless_texture" not in GLConfig.available_extensions:
            return
        
        if not self._meshes_cast_shadows and \
           not self._lines_cast_shadows and \
           not self._points_cast_shadows:
            return

        for spot_light in self.scene.spot_lights:
            if not spot_light.generate_shadows or \
               not spot_light.need_update_depth_map:
                continue

            if spot_light.depth_fbo is None:
                spot_light.depth_fbo = FBO(1024, 1024)
                spot_light.depth_fbo.attach(GL.GL_DEPTH_ATTACHMENT, samplerCube)
                spot_light.depth_fbo.depth_attachment.wrap = GL.GL_CLAMP_TO_EDGE
            
            with GLConfig.LocalConfig(depth_test=True, blend=False, cull_face=None, polygon_mode=GL.GL_FILL):
                with spot_light.depth_fbo:
                    if self._meshes_cast_shadows:
                        self.spot_light_depth_program["spot_light"] = spot_light
                        for mesh, instances in self._meshes_cast_shadows:
                            self.spot_light_depth_program["material"] = mesh.material
                            self.spot_light_depth_program["back_material"] = mesh._back_material
                            self.spot_light_depth_program["explode_distance"] = mesh.explode_distance
                            mesh.draw(self.spot_light_depth_program, instances)

                    if self._lines_cast_shadows:
                        self.spot_light_depth_lines_program["spot_light"] = spot_light
                        for mesh, instances in self._lines_cast_shadows:
                            self.spot_light_depth_lines_program["material"] = mesh.material
                            self.spot_light_depth_lines_program["back_material"] = mesh._back_material
                            mesh.draw(self.spot_light_depth_lines_program, instances)

                    if self._points_cast_shadows:
                        self.spot_light_depth_points_program["spot_light"] = spot_light
                        for mesh, instances in self._points_cast_shadows:
                            self.spot_light_depth_points_program["material"] = mesh.material
                            self.spot_light_depth_points_program["back_material"] = mesh._back_material
                            mesh.draw(self.spot_light_depth_points_program, instances)

            new_handle = spot_light.depth_fbo.depth_attachment.handle
            if spot_light.depth_map_handle != new_handle:
                spot_light.depth_map_handle = new_handle
                self.scene.spot_lights.dirty = True

            spot_light.need_update_depth_map = False

    def update_point_lights_depth(self):
        if "GL_ARB_bindless_texture" not in GLConfig.available_extensions:
            return
        
        if not self._meshes_cast_shadows and \
           not self._lines_cast_shadows and \
           not self._points_cast_shadows:
            return
        
        for point_light in self.scene.point_lights:
            if not point_light.generate_shadows or \
               not point_light.need_update_depth_map:
                continue

            if point_light.depth_fbo is None:
                point_light.depth_fbo = FBO(1024, 1024)
                point_light.depth_fbo.attach(GL.GL_DEPTH_ATTACHMENT, samplerCube)
                point_light.depth_fbo.depth_attachment.wrap = GL.GL_CLAMP_TO_EDGE
            
            with GLConfig.LocalConfig(depth_test=True, blend=False, cull_face=None, polygon_mode=GL.GL_FILL):
                with point_light.depth_fbo:
                    if self._meshes_cast_shadows:
                        self.point_light_depth_program["point_light"] = point_light
                        for mesh, instances in self._meshes_cast_shadows:
                            self.point_light_depth_program["explode_distance"] = mesh.explode_distance
                            self.point_light_depth_program["material"] = mesh.material
                            self.point_light_depth_program["back_material"] = mesh._back_material
                            mesh.draw(self.point_light_depth_program, instances)

                    if self._lines_cast_shadows:
                        self.point_light_depth_lines_program["point_light"] = point_light
                        for mesh, instances in self._lines_cast_shadows:
                            self.point_light_depth_lines_program["material"] = mesh.material
                            self.point_light_depth_lines_program["back_material"] = mesh._back_material
                            mesh.draw(self.point_light_depth_lines_program, instances)

                    if self._points_cast_shadows:
                        self.point_light_depth_points_program["point_light"] = point_light
                        for mesh, instances in self._points_cast_shadows:
                            self.point_light_depth_points_program["material"] = mesh.material
                            self.point_light_depth_points_program["back_material"] = mesh._back_material
                            mesh.draw(self.point_light_depth_points_program, instances)

            new_handle = point_light.depth_fbo.depth_attachment.handle
            if point_light.depth_map_handle != new_handle:
                point_light.depth_map_handle = new_handle
                self.scene.point_lights.dirty = True

            point_light.need_update_depth_map = False

    def update_dir_lights_depth(self):
        if "GL_ARB_bindless_texture" not in GLConfig.available_extensions:
            return
        
        # if not self._meshes_cast_shadows and \
        #    not self._lines_cast_shadows and \
        #    not self._points_cast_shadows:
        #     return
        
        for dir_light in self.scene.dir_lights:
            if not dir_light.generate_shadows:
                continue

            dir_light.max_back_offset = 0

        for mesh, instances in self.scene.all_meshes.items():
            if not mesh.material.cast_shadows:
                continue

            original_center = mesh.center
            original_corner = glm.vec3()
            original_corner.x = original_center.x + (mesh.x_max - mesh.x_min)/2
            original_corner.y = original_center.y + (mesh.y_max - mesh.y_min)/2
            original_corner.z = original_center.z + (mesh.z_max - mesh.z_min)/2
            for instance in instances:
                center = instance.apply(original_center)
                corner = instance.apply(original_corner)
                radius = glm.length(corner - center)

                for dir_light in self.scene.dir_lights:
                    if not dir_light.generate_shadows:
                        continue

                    current_offset = radius + glm.dot(center, -dir_light.direction)
                    if current_offset > dir_light.max_back_offset:
                        dir_light.max_back_offset = current_offset

        for dir_light in self.scene.dir_lights:
            if not dir_light.generate_shadows:
                continue

            if dir_light.depth_fbo is None:
                dir_light.depth_fbo = FBO(1024, 1024, layers=self.camera.CSM_levels)
                dir_light.depth_fbo.attach(GL.GL_DEPTH_ATTACHMENT, sampler2DArray)
                dir_light.depth_fbo.depth_attachment.wrap_s = GL.GL_CLAMP_TO_BORDER
                dir_light.depth_fbo.depth_attachment.wrap_t = GL.GL_CLAMP_TO_BORDER
                dir_light.depth_fbo.depth_attachment.border_color = glm.vec4(1, 1, 1, 1)
            
            with GLConfig.LocalConfig(clear_color=glm.vec4(1,1,1,1), depth_test=True, blend=False, cull_face=None, polygon_mode=GL.GL_FILL):
                with dir_light.depth_fbo:
                    if self._meshes_cast_shadows:
                        self.dir_light_depth_program["dir_light"] = dir_light
                        self.dir_light_depth_program["camera"] = self.camera
                        for mesh, instances in self._meshes_cast_shadows:
                            self.dir_light_depth_program["explode_distance"] = mesh.explode_distance
                            self.dir_light_depth_program["material"] = mesh.material
                            self.dir_light_depth_program["back_material"] = mesh._back_material
                            mesh.draw(self.dir_light_depth_program, instances)

                    if self._lines_cast_shadows:
                        self.dir_light_depth_lines_program["dir_light"] = dir_light
                        self.dir_light_depth_lines_program["camera"] = self.camera
                        for mesh, instances in self._lines_cast_shadows:
                            self.dir_light_depth_lines_program["material"] = mesh.material
                            self.dir_light_depth_lines_program["back_material"] = mesh._back_material
                            mesh.draw(self.dir_light_depth_lines_program, instances)

                    if self._points_cast_shadows:
                        self.dir_light_depth_points_program["dir_light"] = dir_light
                        self.dir_light_depth_points_program["camera"] = self.camera
                        for mesh, instances in self._points_cast_shadows:
                            self.dir_light_depth_points_program["material"] = mesh.material
                            self.dir_light_depth_points_program["back_material"] = mesh._back_material
                            mesh.draw(self.dir_light_depth_points_program, instances)

            new_handle = dir_light.depth_fbo.depth_attachment.handle
            if dir_light.depth_map_handle != new_handle:
                dir_light.depth_map_handle = new_handle
                self.scene.dir_lights.dirty = True
                self._should_update = True

    @property
    def forward_program(self):
        if "forward" in self.programs:
            return self.programs["forward"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        GlassEngineConfig.define_for_program(program)
        program.compile(self_folder + "/../glsl/Pipelines/forward_rendering/forward_rendering.vs")
        program.compile(self_folder + "/../glsl/Pipelines/forward_rendering/forward_rendering.gs")
        program.compile(self_folder + "/../glsl/Pipelines/forward_rendering/forward_rendering.fs")

        program["PointLights"].bind(self.scene.point_lights)
        program["DirLights"].bind(self.scene.dir_lights)
        program["SpotLights"].bind(self.scene.spot_lights)

        self.programs["forward"] = program

        return program
    
    @property
    def forward_lines_program(self):
        if "forward_lines" in self.programs:
            return self.programs["forward_lines"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        GlassEngineConfig.define_for_program(program)
        program.compile(self_folder + "/../glsl/Pipelines/forward_rendering/forward_draw_lines.vs")
        program.compile(self_folder + "/../glsl/Pipelines/forward_rendering/forward_draw_points.fs")

        program["PointLights"].bind(self.scene.point_lights)
        program["DirLights"].bind(self.scene.dir_lights)
        program["SpotLights"].bind(self.scene.spot_lights)

        self.programs["forward_lines"] = program

        return program

    @property
    def forward_points_program(self):
        if "forward_points" in self.programs:
            return self.programs["forward_points"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        GlassEngineConfig.define_for_program(program)
        program.compile(self_folder + "/../glsl/Pipelines/forward_rendering/forward_draw_points.vs")
        program.compile(self_folder + "/../glsl/Pipelines/forward_rendering/forward_draw_points.fs")

        program["PointLights"].bind(self.scene.point_lights)
        program["DirLights"].bind(self.scene.dir_lights)
        program["SpotLights"].bind(self.scene.spot_lights)

        self.programs["forward_points"] = program

        return program

    @property
    def OIT_blend_program(self):
        if "OIT_blend" in self.programs:
            return self.programs["OIT_blend"]
        
        program = ShaderProgram()
        GlassEngineConfig.define_for_program(program)
        program.compile(Frame.draw_frame_vs)
        program.compile(os.path.dirname(os.path.abspath(__file__)) + "/../glsl/Pipelines/OIT_blend.fs")
        self.programs["OIT_blend"] = program
        return program
    
    @property
    def OIT_fbo(self):
        fbo = None
        screen_size = GLConfig.screen_size

        samples = self.camera.screen.samples
        if samples > 1:
            if "OIT_ms" in self.fbos:
                fbo = self.fbos["OIT_ms"]
                fbo.resize(screen_size.x, screen_size.y, samples)
            else:
                fbo = FBO(screen_size.x, screen_size.y, samples)
                fbo.attach(0, sampler2DMS, GL.GL_RGBA32F)
                fbo.attach(1, sampler2DMS, GL.GL_RGBA32F)
                fbo.attach(2, sampler2DMS, GL.GL_R32F)
                fbo.attach(3, sampler2DMS, GL.GL_RGB32F)
                fbo.attach(4, sampler2DMS, GL.GL_RGB32F)
                fbo.attach(GL.GL_DEPTH_ATTACHMENT, sampler2DMS)
                fbo.auto_clear = False
                self.fbos["OIT_ms"] = fbo
            return fbo
        else:
            if "OIT" in self.fbos:
                fbo = self.fbos["OIT"]
                fbo.resize(screen_size.x, screen_size.y)
            else:
                fbo = FBO(screen_size.x, screen_size.y)
                fbo.attach(0, sampler2D, GL.GL_RGBA32F)
                fbo.attach(1, sampler2D, GL.GL_RGBA32F)
                fbo.attach(2, sampler2D, GL.GL_R32F)
                fbo.attach(3, sampler2D, GL.GL_RGB32F)
                fbo.attach(4, sampler2D, GL.GL_RGB32F)
                fbo.attach(GL.GL_DEPTH_ATTACHMENT, sampler2D)
                fbo.auto_clear = False
                self.fbos["OIT"] = fbo
            return fbo

    def env_OIT_blend_fbo(self, instance):
        current_context = GLConfig.buffered_current_context
        key = ("env_OIT_blend_fbo", current_context)
        if key in instance.user_data:
            return instance.user_data[key]
        
        fbo = FBO(2048, 1024)
        fbo.attach(0, sampler2D, GL.GL_RGBA32F)
        fbo.color_attachment(0).filter_mipmap = GL.GL_LINEAR
        instance.user_data[key] = fbo
        return fbo

    def env_FXAA_fbo(self, instance):
        current_context = GLConfig.buffered_current_context
        key = ("env_FXAA_fbo", current_context)
        if key in instance.user_data:
            return instance.user_data[key]
        
        FXAA_fbo = FBO(2048, 1024)
        FXAA_fbo.attach(0, sampler2D)
        FXAA_fbo.color_attachment(0).filter_mipmap = GL.GL_LINEAR
        instance.user_data[key] = FXAA_fbo

        return FXAA_fbo

    @property
    def gen_env_map_program(self):
        if "gen_env_map" not in self.programs:
            self_folder = os.path.dirname(os.path.abspath(__file__))
            program = ShaderProgram()
            GlassEngineConfig.define_for_program(program)
            program.compile(self_folder + "/../glsl/Pipelines/env_mapping/gen_env_map.vs")
            program.compile(self_folder + "/../glsl/Pipelines/env_mapping/gen_env_map.gs")
            program.compile(self_folder + "/../glsl/Pipelines/env_mapping/gen_env_map.fs")
            program["DirLights"].bind(self.scene.dir_lights)
            program["PointLights"].bind(self.scene.point_lights)
            program["SpotLights"].bind(self.scene.spot_lights)
            
            self.programs["gen_env_map"] = program

        return self.programs["gen_env_map"]
    
    @property
    def gen_env_map_lines_program(self):
        if "gen_env_map_lines" not in self.programs:
            self_folder = os.path.dirname(os.path.abspath(__file__))
            program = ShaderProgram()
            GlassEngineConfig.define_for_program(program)
            program.compile(self_folder + "/../glsl/Pipelines/env_mapping/gen_env_map_points.vs")
            program.compile(self_folder + "/../glsl/Pipelines/env_mapping/gen_env_map_lines.gs")
            program.compile(self_folder + "/../glsl/Pipelines/env_mapping/gen_env_map_points.fs")
            program["DirLights"].bind(self.scene.dir_lights)
            program["PointLights"].bind(self.scene.point_lights)
            program["SpotLights"].bind(self.scene.spot_lights)
            
            self.programs["gen_env_map_lines"] = program

        return self.programs["gen_env_map_lines"]

    @property
    def gen_env_map_points_program(self):
        if "gen_env_map_points" not in self.programs:
            self_folder = os.path.dirname(os.path.abspath(__file__))
            program = ShaderProgram()
            GlassEngineConfig.define_for_program(program)
            program.compile(self_folder + "/../glsl/Pipelines/env_mapping/gen_env_map_points.vs")
            program.compile(self_folder + "/../glsl/Pipelines/env_mapping/gen_env_map_points.gs")
            program.compile(self_folder + "/../glsl/Pipelines/env_mapping/gen_env_map_points.fs")
            program["DirLights"].bind(self.scene.dir_lights)
            program["PointLights"].bind(self.scene.point_lights)
            program["SpotLights"].bind(self.scene.spot_lights)
            
            self.programs["gen_env_map_points"] = program

        return self.programs["gen_env_map_points"]
    
    def env_map_fbo(self, instance):
        key = ("env_map_fbo", GLConfig.buffered_current_context)
        if key in instance.user_data:
            return instance.user_data[key]
        
        env_map_fbo = FBO(1024, 1024)
        env_map_fbo.attach(0, samplerCube)
        env_map_fbo.attach(1, samplerCube)
        env_map_fbo.attach(2, samplerCube)
        env_map_fbo.attach(GL.GL_DEPTH_ATTACHMENT, samplerCube)
        env_map_fbo.auto_clear = False
        instance.user_data[key] = env_map_fbo
        
        return env_map_fbo

    def env_bake_times(self, instance):
        key = ("env_bake_times", self.scene)
        if key in instance.user_data:
            return instance.user_data[key]
        
        return 0
    
    def increase_bake_times(self, instance):
        key = ("env_bake_times", self.scene)
        if key not in instance.user_data:
            instance.user_data[key] = 0

        instance.user_data[key] += 1

    def prepare_gen_env_map_draw_mesh(self, view_center:glm.vec3, is_opaque_pass:bool):
        camera = self.camera
        self.gen_env_map_program["CSM_camera"] = camera
        self.gen_env_map_program["view_center"] = view_center
        self.gen_env_map_program["is_opaque_pass"] = is_opaque_pass
        self.gen_env_map_program["background"] = self.scene.background
        if GlassEngineConfig["USE_FOG"]:
            self.gen_env_map_program["fog"] = self.scene.fog

    def gen_env_map_draw_mesh(self, mesh, instances):
        self.gen_env_map_program["explode_distance"] = mesh.explode_distance
        self.gen_env_map_program["material"] = mesh.material
        self.gen_env_map_program["back_material"] = mesh._back_material
        self.gen_env_map_program["is_sphere"] = mesh.is_sphere
        self.gen_env_map_program["mesh_center"] = mesh.center
        mesh.draw(self.gen_env_map_program, instances)

    def prepare_gen_env_map_draw_lines(self, view_center:glm.vec3, is_opaque_pass:bool):
        camera = self.camera
        self.gen_env_map_lines_program["CSM_camera"] = camera
        self.gen_env_map_lines_program["view_center"] = view_center
        self.gen_env_map_lines_program["is_opaque_pass"] = is_opaque_pass
        self.gen_env_map_lines_program["background"] = self.scene.background
        if GlassEngineConfig["USE_FOG"]:
            self.gen_env_map_lines_program["fog"] = self.scene.fog

    def gen_env_map_draw_lines(self, mesh, instances):
        self.gen_env_map_lines_program["material"] = mesh.material
        self.gen_env_map_lines_program["mesh_center"] = mesh.center
        mesh.draw(self.gen_env_map_lines_program, instances)

    def prepare_gen_env_map_draw_points(self, view_center:glm.vec3, is_opaque_pass:bool):
        camera = self.camera
        self.gen_env_map_points_program["CSM_camera"] = camera
        self.gen_env_map_points_program["view_center"] = view_center
        self.gen_env_map_points_program["is_opaque_pass"] = is_opaque_pass
        self.gen_env_map_points_program["background"] = self.scene.background
        if GlassEngineConfig["USE_FOG"]:
            self.gen_env_map_points_program["fog"] = self.scene.fog

    def gen_env_map_draw_points(self, mesh, instances):
        self.gen_env_map_points_program["material"] = mesh.material
        self.gen_env_map_points_program["mesh_center"] = mesh.center
        mesh.draw(self.gen_env_map_points_program, instances)

    def gen_env_map(self, mesh, instances):
        if "GL_ARB_bindless_texture" not in GLConfig.available_extensions:
            return

        max_bake_times = max(mesh.material.env_max_bake_times, mesh._back_material.env_max_bake_times)
        mesh_center = mesh.center
        for instance in instances:
            if instance.visible == 0:
                continue

            if self.env_bake_times(instance) >= max_bake_times:
                continue
            
            view_center = instance.apply(mesh_center)

            instance.visible = 0
            env_map_fbo = self.env_map_fbo(instance)

            with GLConfig.LocalConfig(depth_test=True, depth_write=True, blend=False):
                with env_map_fbo:
                    GLConfig.clear_buffers()

                    if self._opaque_meshes:
                        self.prepare_gen_env_map_draw_mesh(view_center, True)
                        for other_mesh, other_instances in self._opaque_meshes:
                            self.gen_env_map_draw_mesh(other_mesh, other_instances)

                    if self._opaque_lines:
                        self.prepare_gen_env_map_draw_lines(view_center, True)
                        for other_mesh, other_instances in self._opaque_lines:
                            self.gen_env_map_draw_lines(other_mesh, other_instances)

                    if self._opaque_points:
                        self.prepare_gen_env_map_draw_points(view_center, True)
                        for other_mesh, other_instances in self._opaque_points:
                            self.gen_env_map_draw_points(other_mesh, other_instances)

            opaque_color_map = env_map_fbo.color_attachment(0)
            accum_map = None
            reveal_map = None
            if self._transparent_meshes or self._transparent_points or self._transparent_lines:
                with GLConfig.LocalConfig(
                    depth_test=True, depth_write=False, blend=True,
                    blend_src_rgb=GL.GL_ONE, blend_dest_rgb=GL.GL_ONE,
                    blend_src_alpha=GL.GL_ONE, blend_dest_alpha=GL.GL_ONE):
                    GLConfig.blend_src_rgbi[1] = GL.GL_ONE
                    GLConfig.blend_dest_rgbi[1] = GL.GL_ONE
                    GLConfig.blend_src_rgbi[2] = GL.GL_ONE
                    GLConfig.blend_dest_rgbi[2] = GL.GL_ONE
                    with env_map_fbo:
                        GLConfig.clear_buffer(1, glm.vec4(0))
                        GLConfig.clear_buffer(2, glm.vec4(0))

                        if self._transparent_meshes:
                            self.prepare_gen_env_map_draw_mesh(view_center, False)
                            for other_mesh, other_instances in self._transparent_meshes:
                                self.gen_env_map_draw_mesh(other_mesh, other_instances)

                        if self._transparent_lines:
                            self.prepare_gen_env_map_draw_lines(view_center, False)
                            for other_mesh, other_instances in self._transparent_lines:
                                self.gen_env_map_draw_lines(other_mesh, other_instances)

                        if self._transparent_points:
                            self.prepare_gen_env_map_draw_points(view_center, False)
                            for other_mesh, other_instances in self._transparent_points:
                                self.gen_env_map_draw_points(other_mesh, other_instances)

                    # 取出 OIT 信息
                    accum_map = env_map_fbo.color_attachment(1)
                    reveal_map = env_map_fbo.color_attachment(2)

            env_OIT_blend_fbo = self.env_OIT_blend_fbo(instance)
            with GLConfig.LocalConfig(
                polygon_mode=GL.GL_FILL,
                cull_face=None,
                depth_func=GL.GL_ALWAYS
            ):
                with env_OIT_blend_fbo:
                    self.env_OIT_blend_program["opaque_color_map"] = opaque_color_map
                    self.env_OIT_blend_program["accum_map"] = accum_map
                    self.env_OIT_blend_program["reveal_map"] = reveal_map
                    self.env_OIT_blend_program.draw_triangles(Frame.vertices, Frame.indices)

            env_map = env_OIT_blend_fbo.color_attachment(0)
            instance.user_data["env_bake_state"] = "baked"
            env_FXAA_fbo = self.env_FXAA_fbo(instance)
            with env_FXAA_fbo:
                FXAAEffect.program()["screen_image"] = env_map
                FXAAEffect.program().draw_triangles(Frame.vertices, Frame.indices)
            env_map = env_FXAA_fbo.color_attachment(0)
            instance.user_data["env_bake_state"] = "filtered"
            instance.env_map_handle = env_map.handle
            instance.visible = 1
            self.increase_bake_times(instance)

            if self.env_bake_times(instance) <= max_bake_times:
                self._should_update = True

    def prepare_forward_draw_mesh(self, is_opaque_pass:bool):
        camera = self.camera
        self.forward_program["camera"] = camera
        self.forward_program["is_opaque_pass"] = is_opaque_pass
        self.forward_program["background"] = self.scene.background
        if GlassEngineConfig["USE_FOG"]:
            self.forward_program["fog"] = self.scene.fog

    def forward_draw_mesh(self, mesh, instances):
        if mesh.material.need_env_map or mesh._back_material.need_env_map:
            self.gen_env_map(mesh, instances)

        self.forward_program["material"] = mesh.material
        self.forward_program["back_material"] = mesh._back_material
        self.forward_program["explode_distance"] = mesh.explode_distance
        self.forward_program["is_sphere"] = mesh.is_sphere
        self.forward_program["mesh_center"] = mesh.center
        mesh.draw(self.forward_program, instances)

    def prepare_forward_draw_lines(self, is_opaque_pass:bool):
        self.forward_lines_program["camera"] = self.camera
        self.forward_lines_program["is_opaque_pass"] = is_opaque_pass
        self.forward_lines_program["background"] = self.scene.background
        if GlassEngineConfig["USE_FOG"]:
            self.forward_lines_program["fog"] = self.scene.fog

    def forward_draw_lines(self, mesh, instances):
        if mesh.material.need_env_map:
            self.gen_env_map(mesh, instances)

        self.forward_lines_program["material"] = mesh.material
        self.forward_lines_program["mesh_center"] = mesh.center
        mesh.draw(self.forward_lines_program, instances)

    def prepare_forward_draw_points(self, is_opaque_pass:bool):
        self.forward_points_program["camera"] = self.camera
        self.forward_points_program["is_opaque_pass"] = is_opaque_pass
        self.forward_points_program["background"] = self.scene.background
        
        if GlassEngineConfig["USE_FOG"]:
            self.forward_points_program["fog"] = self.scene.fog

    def forward_draw_points(self, mesh, instances):
        if mesh.material.need_env_map:
            self.gen_env_map(mesh, instances)

        self.forward_points_program["material"] = mesh.material
        self.forward_points_program["mesh_center"] = mesh.center
        mesh.draw(self.forward_points_program, instances)

    @property
    def env_OIT_blend_program(self):
        if "env_OIT_blend" in self.programs:
            return self.programs["env_OIT_blend"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        GlassEngineConfig.define_for_program(program)
        program.compile(Frame.draw_frame_vs)
        program.compile(self_folder + "/../glsl/Pipelines/env_mapping/env_OIT_blend.fs")
        self.programs["env_OIT_blend"] = program

        return program

    @property
    def draw_geometry_program(self):
        if "draw_geometry" in self.programs:
            return self.programs["draw_geometry"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        GlassEngineConfig.define_for_program(program)
        program.compile(self_folder + "/../glsl/Pipelines/draw_geometry/draw_geometry.vs")
        program.compile(self_folder + "/../glsl/Pipelines/draw_geometry/draw_geometry.gs")
        program.compile(self_folder + "/../glsl/Pipelines/draw_geometry/draw_geometry.fs")

        self.programs["draw_geometry"] = program

        return program
    
    @property
    def draw_geometry_lines_program(self):
        if "draw_geometry_lines" in self.programs:
            return self.programs["draw_geometry_lines"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        GlassEngineConfig.define_for_program(program)
        program.compile(self_folder + "/../glsl/Pipelines/draw_geometry/draw_geometry_lines.vs")
        program.compile(self_folder + "/../glsl/Pipelines/draw_geometry/draw_geometry_points.fs")
        self.programs["draw_geometry_lines"] = program

        return program
    
    @property
    def draw_geometry_points_program(self):
        if "draw_geometry_points" in self.programs:
            return self.programs["draw_geometry_points"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        GlassEngineConfig.define_for_program(program)
        program.compile(self_folder + "/../glsl/Pipelines/draw_geometry/draw_geometry_points.vs")
        program.compile(self_folder + "/../glsl/Pipelines/draw_geometry/draw_geometry_points.fs")
        self.programs["draw_geometry_points"] = program

        return program

    def draw_geometry(self, mesh, instances):
        self.draw_geometry_program["material"] = mesh.material
        self.draw_geometry_program["back_material"] = mesh._back_material
        self.draw_geometry_program["explode_distance"] = mesh.explode_distance
        mesh.draw(self.draw_geometry_program, instances)

    def draw_geometry_lines(self, mesh, instances):
        self.draw_geometry_lines_program["material"] = mesh.material
        self.draw_geometry_lines_program["back_material"] = mesh._back_material
        self.draw_geometry_lines_program["explode_distance"] = mesh.explode_distance
        mesh.draw(self.draw_geometry_lines_program, instances)

    def draw_geometry_points(self, mesh, instances):
        self.draw_geometry_points_program["material"] = mesh.material
        self.draw_geometry_points_program["back_material"] = mesh._back_material
        self.draw_geometry_points_program["explode_distance"] = mesh.explode_distance
        mesh.draw(self.draw_geometry_points_program, instances)

    def draw_transparent(self):
        if not self._transparent_meshes and \
           not self._transparent_points and \
           not self._transparent_lines:
            return
        
        with GLConfig.LocalConfig(depth_test=True, depth_write=False, blend=True):
            GLConfig.blend_src_rgbi[1] = GL.GL_ONE
            GLConfig.blend_dest_rgbi[1] = GL.GL_ONE
            GLConfig.blend_src_rgbi[2] = GL.GL_ONE
            GLConfig.blend_dest_rgbi[2] = GL.GL_ONE
            with self.OIT_fbo:
                GLConfig.clear_buffer(1, glm.vec4(0))
                GLConfig.clear_buffer(2, glm.vec4(0))

                if self._transparent_meshes:
                    self.prepare_forward_draw_mesh(False)
                    for mesh, instances in self._transparent_meshes:
                        self.forward_draw_mesh(mesh, instances)

                if self._transparent_lines:
                    self.prepare_forward_draw_lines(False)
                    for mesh, instances in self._transparent_lines:
                        self.forward_draw_lines(mesh, instances)

                if self._transparent_points:
                    self.prepare_forward_draw_points(False)
                    for mesh, instances in self._transparent_points:
                        self.forward_draw_points(mesh, instances)

        # 取出 OIT 信息
        resolved = self.OIT_fbo.resolved
        accum_map = resolved.color_attachment(1)
        reveal_map = resolved.color_attachment(2)

        with GLConfig.LocalConfig(
            polygon_mode=GL.GL_FILL,
            cull_face=None,
            depth_func=GL.GL_ALWAYS,
            blend=True,
            blend_func=(GL.GL_ONE_MINUS_SRC_ALPHA, GL.GL_SRC_ALPHA)
        ):
            self.OIT_blend_program["accum_map"] = accum_map
            self.OIT_blend_program["reveal_map"] = reveal_map
            self.OIT_blend_program.draw_triangles(Frame.vertices, Frame.indices)

        if self.camera.screen.post_process_effects.need_pos_info:
            used_fbo = None
            if self.__class__.__name__ == "ForwardRenderer":
                used_fbo = self.OIT_fbo
            elif self.__class__.__name__ == "DeferredRenderer":
                used_fbo = self.gbuffer

            with GLConfig.LocalConfig(depth_test=True, depth_write=True, blend=False):
                with used_fbo:
                    if self._transparent_meshes:
                        self.draw_geometry_program["camera"] = self.camera
                        for mesh, instances in self._transparent_meshes:
                            self.draw_geometry(mesh, instances)

                    if self._transparent_lines:
                        self.draw_geometry_lines_program["camera"] = self.camera
                        for mesh, instances in self._transparent_lines:
                            self.draw_geometry_lines(mesh, instances)

                    if self._transparent_points:
                        self.draw_geometry_points_program["camera"] = self.camera
                        for mesh, instances in self._transparent_points:
                            self.draw_geometry_points(mesh, instances)

            resolved = used_fbo.resolved
            self._depth_map = resolved.depth_attachment
            self._view_pos_map = resolved.color_attachment(3)
            self._view_normal_map = resolved.color_attachment(4)



from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm
import math

def init_Hexahedron(cls):
    cls.edge_length = 2/3*math.sqrt(3)
    cls.base_positions = \
    [
        glm.normalize(glm.vec3(-1, -1, -1)), # 0
        glm.normalize(glm.vec3( 1, -1, -1)), # 1
        glm.normalize(glm.vec3( 1,  1, -1)), # 2
        glm.normalize(glm.vec3(-1,  1, -1)), # 3

        glm.normalize(glm.vec3(-1, -1,  1)), # 4
        glm.normalize(glm.vec3( 1, -1,  1)), # 5
        glm.normalize(glm.vec3( 1,  1,  1)), # 6
        glm.normalize(glm.vec3(-1,  1,  1)), # 7
    ]

    cls.base_indices = \
    [
        # 下面
        glm.uvec3(3, 2, 1),
        glm.uvec3(3, 1, 0),

        # 上面
        glm.uvec3(4, 5, 6),
        glm.uvec3(4, 6, 7),

        # 左面
        glm.uvec3(3, 0, 4),
        glm.uvec3(3, 4, 7),

        # 右面
        glm.uvec3(1, 2, 6),
        glm.uvec3(1, 6, 5),

        # 前面
        glm.uvec3(0, 1, 5),
        glm.uvec3(0, 5, 4),

        # 后面
        glm.uvec3(2, 3, 7),
        glm.uvec3(2, 7, 6)
    ]

    cls.fixed_vertices = []
    for i, index in enumerate(cls.base_indices):
        pos0 = cls.base_positions[index[0]]
        pos1 = cls.base_positions[index[1]]
        pos2 = cls.base_positions[index[2]]

        tex_coord0 = glm.vec3(0)
        tex_coord1 = glm.vec3(1, 0, 0)
        tex_coord2 = glm.vec3(1, 1, 0)
        if i % 2 == 1:
            tex_coord0 = glm.vec3(0)
            tex_coord1 = glm.vec3(1, 1, 0)
            tex_coord2 = glm.vec3(0, 1, 0)

        v1 = pos1 - pos0
        v2 = pos2 - pos0
        normal = glm.normalize(glm.cross(v1, v2))

        vertex0 = Vertex(
            position=pos0,
            normal=normal,
            tex_coord=tex_coord0
        )
        vertex1 = Vertex(
            position=pos1,
            normal=normal,
            tex_coord=tex_coord1
        )
        vertex2 = Vertex(
            position=pos2,
            normal=normal,
            tex_coord=tex_coord2
        )
        cls.fixed_vertices.append(vertex0)
        cls.fixed_vertices.append(vertex1)
        cls.fixed_vertices.append(vertex2)

    return cls

@init_Hexahedron
class Hexahedron(Mesh):

    @checktype
    def __init__(self, radius=1,
                 color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), back_color:(glm.vec3,glm.vec4)=None,
                 normalize_tex_coord=False,
                 name:str=""):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=True)
        self.__radius = radius
        self.__normalize_tex_coord = normalize_tex_coord
        self.start_building()

    def build(self):
        self.is_closed = True
        self.self_calculated_normal = True

        vertices = self.vertices
        indices = self.indices
        radius = self.__radius
        normalize_tex_coord = self.__normalize_tex_coord

        for i, fix_vert in enumerate(Hexahedron.fixed_vertices):
            tex_coord = fix_vert.tex_coord
            if not normalize_tex_coord:
                tex_coord = tex_coord * Hexahedron.edge_length
            
            vertices[i] = Vertex(
                position = radius * fix_vert.position,
                normal = fix_vert.normal,
                tex_coord = tex_coord
            )

        for i in range(0, len(Hexahedron.fixed_vertices), 3):
            indices[int(i/3)] = glm.uvec3(i, i+1, i+2)
            self.generate_temp_TBN(vertices[i], vertices[i+1], vertices[i+2])

    @property
    def radius(self):
        return self.__radius
    
    @radius.setter
    @Mesh.param_setter
    def radius(self, radius:float):
        self.__radius = radius

    @property
    def normalize_tex_coord(self):
        return self.__normalize_tex_coord
    
    @normalize_tex_coord.setter
    @Mesh.param_setter
    def normalize_tex_coord(self, flag:bool):
        self.__normalize_tex_coord = flag


from .Block import Block
from .SSBO import SSBO

from .utils import checktype
from functools import wraps

class ShaderStorageBlock(Block):
	
	_bound_vars = {}
	BO = SSBO

	class HostClass:
		def __init__(self):
			self._dirty = True

		@property
		def dirty(self):
			return self._dirty
		
		@dirty.setter
		@checktype
		def dirty(self, flag:bool):
			self._dirty = flag

		def upload(self):
			if self._dirty:
				ShaderStorageBlock.upload_var(self)
				self._dirty = False

		@staticmethod
		def not_const(func):
			@wraps(func)
			def wrapper(*args, **kwargs):
				self = args[0]
				result = func(*args, **kwargs)
				self._dirty = True
				return result
			
			return wrapper

	def __init__(self, program):
		Block.__init__(self, program)



from ..Mesh import Mesh

from glass import Vertex

import glm
import math

class TrefoilKnot(Mesh):

    def __init__(self, tube_radius:float=0.2, knot_radius:float=1,
                 color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), back_color:(glm.vec3,glm.vec4)=None,
                 vertical:bool=False, normalize_tex_coord=False,
                 n_lon_divide:int=200, start_lon:float=0, span_lon:float=360,
                 n_lat_divide:int=36, start_lat:float=0, span_lat:float=360,
                 name:str="", block:bool=True):
        
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=block)
        self.__tube_radius = tube_radius
        self.__knot_radius = knot_radius
        self.__vertical = vertical
        self.__normalize_tex_coord = normalize_tex_coord
        self.__n_lon_divide = n_lon_divide
        self.__start_lon = start_lon
        self.__span_lon = span_lon
        self.__n_lat_divide = n_lat_divide
        self.__start_lat = start_lat
        self.__span_lat = span_lat
        self.start_building()

    def build(self):
        self.is_closed = True
        self.self_calculated_normal = True

        vertices = self.vertices
        indices = self.indices
        r = self.__tube_radius
        R = self.__knot_radius/3
        vertical = self.__vertical
        normalize_tex_coord = self.__normalize_tex_coord
        n_lon_divide = self.__n_lon_divide
        start_lon = self.__start_lon/180*math.pi
        span_lon = self.__span_lon/180*math.pi
        n_lat_divide = self.__n_lat_divide
        start_lat = self.__start_lat/180*math.pi
        span_lat = self.__span_lat/180*math.pi

        i_vertex = 0
        i_index = 0

        L = 0
        last_normalized_center = None
        for i in range(n_lon_divide):
            theta = start_lon + span_lon*i/(n_lon_divide-1)
            sin_theta = math.sin(theta)
            cos_theta = math.cos(theta)
            sin_2theta = math.sin(2*theta)
            cos_2theta = math.cos(2*theta)
            sin_3theta = math.sin(3*theta)
            cos_3theta = math.cos(3*theta)

            x0 = sin_theta + 2*sin_2theta
            y0 = cos_theta - 2*cos_2theta
            z0 = -sin_3theta
            normalized_center = glm.vec3(x0, y0, z0)

            if i > 0:
                L += glm.length(normalized_center - last_normalized_center)

            last_normalized_center = normalized_center

            x0 *= R
            y0 *= R
            z0 *= R

            dx = cos_theta + 4*cos_2theta
            dy = -sin_theta + 4*sin_2theta
            dz = -3*cos_3theta

            ddx = -sin_theta - 8*sin_2theta
            ddy = -cos_theta + 8*cos_2theta
            ddz = 9*sin_3theta

            alpha = glm.normalize(glm.vec3(dx, dy, dz))
            gamma = glm.normalize(glm.cross(glm.vec3(dx, dy, dz), glm.vec3(ddx, ddy, ddz)))
            beta = glm.cross(gamma, alpha)

            for j in range(n_lat_divide):
                phi = start_lat + span_lat*j/(n_lat_divide-1)

                normal = math.cos(phi) * gamma + math.sin(phi) * beta
                x = x0 + r*normal.x
                y = y0 + r*normal.y
                z = z0 + r*normal.z
                s = L
                t = r*phi
                if normalize_tex_coord:
                    t = t / r / (2*math.pi)
                    s = s / r / (2*math.pi)
                
                vertex = Vertex()
                vertex.position = glm.vec3(x, y, z)
                vertex.normal = normal
                vertex.tex_coord = glm.vec3(s, t, 0)

                if vertical:
                    vertex.position = glm.vec3(vertex.position.x, -vertex.position.z, vertex.position.y)
                    vertex.normal = glm.vec3(vertex.normal.x, -vertex.normal.z, vertex.normal.y)

                vertices[i_vertex] = vertex
                i_vertex += 1

                if i > 0 and j > 0:
                    triangle = glm.uvec3(0, 0, 0)
                    triangle[0] = i_vertex - 1
                    triangle[1] = i_vertex - 1 - 1 - n_lat_divide
                    triangle[2] = i_vertex - 1 - 1
                    indices[i_index] = triangle
                    i_index += 1
                    self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                    triangle = glm.uvec3(0, 0, 0)
                    triangle[0] = i_vertex - 1
                    triangle[1] = i_vertex - 1 - n_lat_divide
                    triangle[2] = i_vertex - 1 - 1 - n_lat_divide
                    indices[i_index] = triangle
                    i_index += 1
                    self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])
            
                    yield

        del vertices[i_vertex:]
        del indices[i_index:]

    @property
    def tube_radius(self):
        return self.__tube_radius
    
    @tube_radius.setter
    @Mesh.param_setter
    def tube_radius(self, radius:float):
        self.__tube_radius = radius

    @property
    def knot_radius(self):
        return self.__knot_radius
    
    @knot_radius.setter
    @Mesh.param_setter
    def knot_radius(self, radius:float):
        self.__knot_radius = radius

    @property
    def vertical(self):
        return self.__vertical
    
    @vertical.setter
    @Mesh.param_setter
    def vertical(self, flag:bool):
        self.__vertical = flag

    @property
    def normalize_tex_coord(self):
        return self.__normalize_tex_coord
    
    @normalize_tex_coord.setter
    @Mesh.param_setter
    def normalize_tex_coord(self, flag:bool):
        self.__normalize_tex_coord = flag

    @property
    def n_lon_divide(self):
        return self.__n_lon_divide
    
    @n_lon_divide.setter
    @Mesh.param_setter
    def n_lon_divide(self, n:int):
        self.__n_lon_divide = n

    @property
    def start_lon(self):
        return self.__start_lon
    
    @start_lon.setter
    @Mesh.param_setter
    def start_lon(self, angle_deg:float):
        self.__start_lon = angle_deg

    @property
    def span_lon(self):
        return self.__span_lon
    
    @span_lon.setter
    @Mesh.param_setter
    def span_lon(self, angle_deg:float):
        self.__span_lon = angle_deg

    @property
    def n_lat_divide(self):
        return self.__n_lat_divide
    
    @n_lat_divide.setter
    @Mesh.param_setter
    def n_lat_divide(self, n:int):
        self.__n_lat_divide = n

    @property
    def start_lat(self):
        return self.__start_lat
    
    @start_lat.setter
    @Mesh.param_setter
    def start_lat(self, angle_deg:float):
        self.__start_lat = angle_deg

    @property
    def span_lat(self):
        return self.__span_lat
    
    @span_lat.setter
    @Mesh.param_setter
    def span_lat(self, angle_deg:float):
        self.__span_lat = angle_deg


import glm
import math
import copy
import numpy as np
from glass.AttrList import AttrList

def fzero(f, interval):
    lower = interval[0]
    upper = interval[1]
    f_lower = f(lower)
    
    f_upper = None
    if math.isinf(upper):
        d = 1
        upper = lower + d
        f_upper = f(upper)
        times = 0
        while np.sign(f_lower) == np.sign(f_upper):
            d *= 2
            upper = lower + d
            f_upper = f(upper)
            times += 1
            if times > 20:
                return None
    else:
        f_upper = f(upper)
    
    if np.sign(f_lower) == np.sign(f_upper):
        return None
    
    while upper - lower > 1E-6:
        middle = (lower + upper)/2
        f_middle = f(middle)
        if np.sign(f_middle) == np.sign(f_lower):
            f_lower = f_middle
            lower = middle
        else:
            f_upper = f_middle
            upper = middle

    return (lower + upper)/2

def angle_of(v1, v2):
    len_v1 = math.sqrt(v1[0]**2 + v1[1]**2 + v1[2]**2)
    len_v2 = math.sqrt(v2[0]**2 + v2[1]**2 + v2[2]**2)
    if len_v1 < 1E-6 or len_v2 < 1E-6:
        return 0

    cos_theta = (v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2]) / (len_v1 * len_v2)
    if cos_theta < -1:
        cos_theta = -1
    if cos_theta > 1:
        cos_theta = 1

    return math.acos(cos_theta)

def cos_angle_of(v1, v2):
    len_v1 = glm.length(v1)
    len_v2 = glm.length(v2)
    if len_v1 < 1E-6 or len_v2 < 1E-6:
        return 0

    cos_theta = glm.dot(v1, v2) / (len_v1 * len_v2)
    if cos_theta < -1:
        cos_theta = -1
    if cos_theta > 1:
        cos_theta = 1
        
    return cos_theta

def approx_pos(pos, precision:int=7):
    return glm.vec3(round(pos.x, precision), round(pos.y, precision), round(pos.z, precision))

def hash_pos(pos):
    negtive_zero = "-0.0000000"
    zero = "0.0000000"

    x_str = f"{pos[0]:.7f}"
    if x_str == negtive_zero:
        x_str = zero

    y_str = f"{pos[1]:.7f}"
    if y_str == negtive_zero:
        y_str = zero

    z_str = f"{pos[2]:.7f}"
    if z_str == negtive_zero:
        z_str = zero

    return f"({x_str}, {y_str}, {z_str})"

def bounding_box(vertices):
    x_min = float("inf")
    x_max = -float("inf")
    y_min = float("inf")
    y_max = -float("inf")
    z_min = float("inf")
    z_max = -float("inf")

    if not vertices:
        return (x_min, x_max, y_min, y_max, z_min, z_max)
    
    for vertex in vertices:
        if vertex.position.x < x_min:
            x_min = vertex.position.x
        if vertex.position.x > x_max:
            x_max = vertex.position.x

        if vertex.position.y < y_min:
            y_min = vertex.position.y
        if vertex.position.y > y_max:
            y_max = vertex.position.y

        if vertex.position.z < z_min:
            z_min = vertex.position.z
        if vertex.position.z > z_max:
            z_max = vertex.position.z

    return (x_min, x_max, y_min, y_max, z_min, z_max)

def is_closed(vertices, indices):
    if not indices:
        return False
    
    points_map = {}
    for index in indices:
        vertex0 = vertices[index[0]]
        vertex1 = vertices[index[1]]
        vertex2 = vertices[index[2]]

        hash_pos0 = hash_pos(vertex0.position)
        hash_pos1 = hash_pos(vertex1.position)
        hash_pos2 = hash_pos(vertex2.position)

        v01 = glm.normalize(vertex1.position - vertex0.position)
        v10 = -v01
        v02 = glm.normalize(vertex2.position - vertex0.position)
        v20 = -v02
        v12 = glm.normalize(vertex2.position - vertex1.position)
        v21 = -v12

        hash_v01 = hash_pos(v01)
        hash_v10 = hash_pos(v10)
        hash_v02 = hash_pos(v02)
        hash_v20 = hash_pos(v20)
        hash_v12 = hash_pos(v12)
        hash_v21 = hash_pos(v21)

        # 0
        if hash_pos0 not in points_map:
            points_map[hash_pos0] = set()

        if hash_v01 not in points_map[hash_pos0]:
            points_map[hash_pos0].add(hash_v01)
        else:
            points_map[hash_pos0].remove(hash_v01)
            if not points_map[hash_pos0]:
                del points_map[hash_pos0]

        if hash_v02 not in points_map[hash_pos0]:
            points_map[hash_pos0].add(hash_v02)
        else:
            points_map[hash_pos0].remove(hash_v02)
            if not points_map[hash_pos0]:
                del points_map[hash_pos0]

        # 1
        if hash_pos1 not in points_map:
            points_map[hash_pos1] = set()

        if hash_v10 not in points_map[hash_pos1]:
            points_map[hash_pos1].add(hash_v10)
        else:
            points_map[hash_pos1].remove(hash_v10)
            if not points_map[hash_pos1]:
                del points_map[hash_pos1]

        if hash_v12 not in points_map[hash_pos1]:
            points_map[hash_pos1].add(hash_v12)
        else:
            points_map[hash_pos1].remove(hash_v12)
            if not points_map[hash_pos1]:
                del points_map[hash_pos1]

        # 2
        if hash_pos2 not in points_map:
            points_map[hash_pos2] = set()

        if hash_v20 not in points_map[hash_pos2]:
            points_map[hash_pos2].add(hash_v20)
        else:
            points_map[hash_pos2].remove(hash_v20)
            if not points_map[hash_pos2]:
                del points_map[hash_pos2]

        if hash_v21 not in points_map[hash_pos2]:
            points_map[hash_pos2].add(hash_v21)
        else:
            points_map[hash_pos2].remove(hash_v21)
            if not points_map[hash_pos2]:
                del points_map[hash_pos2]
    
    return (not points_map)

def normalize(arr):
    if len(arr.shape) == 2:
        lens = np.sqrt( arr[:,0]**2 + arr[:,1]**2 + arr[:,2]**2 )
        good_lens_it = (lens > 1E-6)

        normalized_arr = arr[:]
        normalized_arr[:,0][good_lens_it] /= lens[good_lens_it]
        normalized_arr[:,1][good_lens_it] /= lens[good_lens_it]
        normalized_arr[:,2][good_lens_it] /= lens[good_lens_it]
        return normalized_arr
    else:
        len_arr = np.linalg.norm(arr)
        if len_arr > 1E-6:
            return arr / len_arr
        else:
            return arr[:]

def generate_TBN(vertices, indices):
    position_view = vertices["position"].ndarray[indices.ndarray.flat, :]
    position0_view = position_view[0::3, :]
    position1_view = position_view[1::3, :]
    position2_view = position_view[2::3, :]

    tex_coord_view = vertices["tex_coord"].ndarray[indices.ndarray.flat, :]
    tex_coord0_view = tex_coord_view[0::3, :]
    tex_coord1_view = tex_coord_view[1::3, :]
    tex_coord2_view = tex_coord_view[2::3, :]

    v01 = position1_view - position0_view
    v02 = position2_view - position0_view

    st01 = tex_coord1_view - tex_coord0_view
    st02 = tex_coord2_view - tex_coord0_view

    det = st01[:,0] * st02[:,1] - st02[:,0] * st01[:,0]
    good_det_id = (abs(det) > 1E-6)
    det = det.reshape(st01.shape[0], 1)

    normal = normalize(np.cross(v01, v02))
    tangent = st02[:,1].reshape(v01.shape[0], 1)*v01 - st01[:,1].reshape(v01.shape[0], 1)*v02
    bitangent = st01[:,0].reshape(v01.shape[0], 1)*v02 - st02[:,0].reshape(v01.shape[0], 1)*v01
    
    tangent[good_det_id,:] /= det[good_det_id,:]
    bitangent[good_det_id,:] /= det[good_det_id,:]

    return tangent, bitangent, normal

def generate_auto_TBN(vertices, indices, calculate_normal=True):
    if "tangent" not in vertices._attr_list_map:
        vertices._attr_list_map["tangent"] = AttrList(dtype=glm.vec3)

    if "bitangent" not in vertices._attr_list_map:
        vertices._attr_list_map["bitangent"] = AttrList(dtype=glm.vec3)

    if "normal" not in vertices._attr_list_map:
        vertices._attr_list_map["normal"] = AttrList(dtype=glm.vec3)

    points_list = []
    for i in range(vertices["position"].ndarray.shape[0]):
        points_list.append({
            "tangent": np.array([0, 0, 0], dtype=np.float32),
            "bitangent": np.array([0, 0, 0], dtype=np.float32),
            "normal": np.array([0, 0, 0], dtype=np.float32),
            "num": 0,
            "index": set()
        })

    tangent, bitangent, normal = generate_TBN(vertices, indices)

    for i in range(len(indices.ndarray.flat)):
        index = indices.ndarray.flat[i]
        info = points_list[index]
        info["num"] += 1
        info["tangent"] += tangent[i//3, :]
        info["bitangent"] += bitangent[i//3, :]
        info["normal"] += normal[i//3, :]
        info["index"].add(index)
                
    for info in points_list:
        if info["num"] == 0:
            continue

        info["tangent"] = info["tangent"] / info["num"]
        info["bitangent"] = info["bitangent"] / info["num"]
        len_normal = np.linalg.norm(info["normal"])
        if len_normal > 1E-6:
            info["normal"] = info["normal"] / len_normal

        for index in info["index"]:
            vertices["tangent"].ndarray[index, :] = info["tangent"]
            vertices["bitangent"].ndarray[index, :] = info["bitangent"]
            if calculate_normal:
                vertices["normal"].ndarray[index, :] = info["normal"]

def generate_smooth_TBN(vertices, indices, calculate_normal=True):
    if "tangent" not in vertices._attr_list_map:
        vertices._attr_list_map["tangent"] = AttrList(dtype=glm.vec3)

    if "bitangent" not in vertices._attr_list_map:
        vertices._attr_list_map["bitangent"] = AttrList(dtype=glm.vec3)

    if "normal" not in vertices._attr_list_map:
        vertices._attr_list_map["normal"] = AttrList(dtype=glm.vec3)

    points_map = {}
    pos_keys = []
    for i in range(vertices["position"].ndarray.shape[0]):
        pos = vertices["position"].ndarray[i]
        pos_key = hash_pos(pos)
        pos_keys.append(pos_key)
        if pos_key not in points_map:
            points_map[pos_key] = \
            {
                "tangent": np.array([0, 0, 0], dtype=np.float32),
                "bitangent": np.array([0, 0, 0], dtype=np.float32),
                "normal": np.array([0, 0, 0], dtype=np.float32),
                "num": 0,
                "index": set()
            }

    tangent, bitangent, normal = generate_TBN(vertices, indices)

    for i in range(len(indices.ndarray.flat)):
        index = indices.ndarray.flat[i]
        pos_key = pos_keys[index]
        info = points_map[pos_key]
        info["num"] += 1
        info["tangent"] += tangent[i//3, :]
        info["bitangent"] += bitangent[i//3, :]
        info["normal"] += normal[i//3, :]
        info["index"].add(index)
                
    for info in points_map.values():
        if info["num"] == 0:
            continue

        info["tangent"] = info["tangent"] / info["num"]
        info["bitangent"] = info["bitangent"] / info["num"]
        len_normal = np.linalg.norm(info["normal"])
        if len_normal > 1E-6:
            info["normal"] = info["normal"] / len_normal

        for index in info["index"]:
            vertices["tangent"].ndarray[index, :] = info["tangent"]
            vertices["bitangent"].ndarray[index, :] = info["bitangent"]
            if calculate_normal:
                vertices["normal"].ndarray[index, :] = info["normal"]
        
def line_intersect_plane(line_start:glm.vec3, line_direction:glm.vec3,
                         plane_start:glm.vec3, plane_normal:glm.vec3):
    A = plane_start
    B = line_start
    d = line_direction
    n = plane_normal
    AB = B - A

    return B - glm.dot(AB, n)/glm.dot(d, n) * d

def point_distance_to_line(point:glm.vec3, line_start:glm.vec3, line_direction:glm.vec3):
    A = line_start
    d = line_direction
    B = point

    AB = B - A
    return glm.length(glm.cross(AB, d))

def point_mirror_by_plane(point:glm.vec3, plane_start:glm.vec3, plane_normal:glm.vec3):
    d = plane_start - point
    return point + 2 * glm.dot(d, plane_normal) * plane_normal

def polygon_centroid(polygon):
    polygon = copy.copy(polygon)

    O = None
    for point in polygon:
        if O is None:
            O = point
        else:
            O = O + point

    len_polygon = len(polygon)
    O /= len_polygon

    centroid = None
    weight_sum = 0
    for i in range(1, len_polygon+1):
        point0 = polygon[i-1]
        point1 = polygon[i] if i < len_polygon else polygon[0]
        center = (O + point0 + point1) / 3
        weight = glm.length(glm.cross(point0-O, point1-O))/2
        weight_sum += weight
        if centroid is None:
            centroid = weight * center
        else:
            centroid += weight * center

    return centroid / weight_sum

def polygon_normal(polygon, centroid=None):
    if centroid is None:
        centroid = polygon_centroid(polygon)

    normal = glm.vec3(0)
    weight_sum = 0
    len_polygon = len(polygon)
    for i in range(1, len_polygon+1):
        point0 = polygon[i-1]
        point1 = polygon[i] if i < len_polygon else polygon[0]

        v0 = point0 - centroid
        v1 = point1 - centroid

        n = glm.cross(v0, v1)
        len_n = glm.length(n)
        if len_n < 1E-6:
            continue
        n /= len_n

        weight = angle_of(v0, v1)
        weight_sum += weight
        normal += weight * n

    return glm.normalize(normal)

def is_even(num:int):
    return num % 2 == 0

def is_odd(num:int):
    return num % 2 != 0

def factorial(n:int):
    return np.math.factorial(n)

def Zernike_coeff(n:int, m:int, k:int):
    sgn = 1 if is_even(k) else -1
    C = factorial(n - k) / (factorial(k) * factorial((n+m)//2 - k) * factorial((n-m)//2 - k))
    return sgn * C

def Zernike_eval(n:int, m:int, r:float, theta:float):
    use_cos = (m >= 0)
    m = abs(m)

    if n < 0 or is_odd(n - m):
        return 0
    
    R = 0
    ub = (n - m) // 2
    for k in range(ub+1):
        R += Zernike_coeff(n, m, k) * r**(n-2*k)

    if use_cos:
        return R * np.cos(m*theta)
    else:
        return R * np.sin(m*theta)

def associated_Legendre_coeff(l:int, m:int, k:int):
    sgn = 1 if is_even(k) else -1
    C = factorial(2*(l - k)) / (2**l * factorial(k) * factorial(l-k) * factorial(l - 2*k - m))
    return sgn * C

def associated_Legendre_eval(l:int, m:int, x:float):
    if l < 0 or l < abs(m):
        return 0

    if m < 0:
        m = abs(m)
        sgn = 1 if is_even(m) else -1
        C = factorial(l - m) / factorial(l + m)
        return sgn * C * associated_Legendre_eval(l, m, x)

    P = 0
    ub = (l - m) // 2
    for k in range(ub + 1):
        P += associated_Legendre_coeff(l, m, k) * x**(l - 2*k - m)

    P *= (1 - x**2)**(m/2)
    return P

def spherical_harmonics_coeff(l:int, m:int):
    sgn = 1 if is_even(m) else -1
    return sgn * math.sqrt((2*l+1)/(4*math.pi) * factorial(l-m)/factorial(l+m))

def spherical_harmonics_eval(l:int, m:int, theta:float, phi:float):
    A = spherical_harmonics_coeff(l, m)
    P = associated_Legendre_eval(l, m, np.cos(theta))
    return A * P * np.exp(m*phi*1j)
    


from .WeakSet import WeakSet
import functools

class MetaInstancesRecorder(type):

    _all_instances = {}
    
    @property
    def all_instances(cls):
        if cls not in MetaInstancesRecorder._all_instances:
            MetaInstancesRecorder._all_instances[cls] = WeakSet()

        return MetaInstancesRecorder._all_instances[cls]
    
    @staticmethod
    def init(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            self = args[0]
            if self.__class__ not in MetaInstancesRecorder._all_instances:
                MetaInstancesRecorder._all_instances[self.__class__] = WeakSet()

            MetaInstancesRecorder._all_instances[self.__class__].add(self)
            return func(*args, **kwargs)

        return wrapper
    
    @staticmethod
    def delete(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            self = args[0]
            if self.__class__ in MetaInstancesRecorder._all_instances:
                MetaInstancesRecorder._all_instances[self.__class__].remove(self)
            
            return func(*args, **kwargs)

        return wrapper


from OpenGL import GL

from .SSUBO import SSUBO
from .utils import subscript_set, subscript

class SSBO(SSUBO):

	_basic_info = \
	{
		"gen_func": GL.glGenBuffers,
		"bind_func": GL.glBindBuffer,
		"del_func": GL.glDeleteBuffers,
		"target_type": GL.GL_SHADER_STORAGE_BUFFER,
		"binding_type": GL.GL_SHADER_STORAGE_BUFFER_BINDING,
		"need_number": True,
	}

	_binding_points_pool = None
	_current_context = 0

	def __init__(self):
		SSUBO.__init__(self)

	def read_back(self, start:int, nbytes:int)->bytearray:
		if start < 0:
			raise ValueError("Memory start position should be positive, " + str(start) + " is passed")
		
		if nbytes < 0:
			raise ValueError("'nbytes' should be positive, " + str(nbytes) + " is passed")

		if start >= self._nbytes:
			raise ValueError("Memory start position is out of range, max position is " + str(self._nbytes-1) + ", " + str(start) + " is passed.")

		if start + nbytes > self._nbytes:
			raise ValueError("Memory end position is out of range, max position is " + str(self._nbytes) +  ", " + str(start + nbytes) + " is applied.")

		if nbytes == 0:
			return bytearray()

		self.bind()
		ptr_data = GL.glMapBuffer(GL.GL_SHADER_STORAGE_BUFFER, GL.GL_READ_ONLY)
		data = (GL.GLubyte*nbytes).from_address(ptr_data)
		self._buffer[start:start+nbytes] = data[:]
		GL.glUnmapBuffer(GL.GL_SHADER_STORAGE_BUFFER)
		
		return self._buffer[start:start+nbytes]
	
	def download(self):
		len_array = None
		for atom_name, atom_info in self._atom_info_map.items():
			atom_type = atom_info["type"]
			get_func = SSUBO._get_atom_func(atom_type)
			atom_offset = atom_info["offset"]
			subscript_chain = atom_info["subscript_chain"]
			if "[{0}]" not in atom_name:
				value = get_func(self, atom_offset)
				subscript_set(self._bound_var, subscript_chain, value)
			else:
				if len_array is None:
					# pos_array_end = atom_name.find("[{0}]")
					# len_array = len(eval("self._bound_var." + atom_name[:pos_array_end]))
					
					pos_array_end = subscript_chain.index(("getitem", "{0}"))
					variable_length_array = subscript(self._bound_var, subscript_chain[:pos_array_end])
					len_array = len(variable_length_array)

				stride = atom_info["stride"]
				for i in range(len_array):
					value = get_func(self, atom_offset + i*stride)
					subscript_set(self._bound_var, subscript_chain, value, i)



from PyQt6.QtWidgets import QWidget, QHBoxLayout, QLabel, QDoubleSpinBox, QSlider, QCheckBox
from PyQt6.QtCore import pyqtSignal, Qt

class SliderInput(QWidget):

    checked_changed = pyqtSignal(bool)
    value_changed = pyqtSignal(float)

    def __init__(self, range:tuple, value:float=None, prompt:str=None, checkable:bool=False, unit:str=None, parent:QWidget=None):
        QWidget.__init__(self, parent=parent)
        self.should_call_slot = True

        hlayout = QHBoxLayout()
        hlayout.setAlignment(Qt.AlignmentFlag.AlignLeft)

        if prompt:
            if not checkable:
                self.label = QLabel(prompt, parent=self)
            else:
                self.label = QCheckBox(prompt, parent=self)
                self.label.stateChanged.connect(self.slot_state_changed)
            hlayout.addWidget(self.label)

        self.spinbox = QDoubleSpinBox(parent=self)
        if unit:
            self.spinbox.setSuffix(unit)
        self.spinbox.setRange(range[0], range[1])
        hlayout.addWidget(self.spinbox)

        self.slider = QSlider(Qt.Orientation.Horizontal, parent=self)
        self.slider.setRange(0, 1000)
        hlayout.addWidget(self.slider)

        hlayout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(hlayout)

        if value is not None:
            self.spinbox.setValue(value)
            default_tick = self.value_to_tick(value)
            self.slider.setValue(default_tick)

        self.spinbox.valueChanged.connect(self.spinbox_value_changed)
        self.slider.valueChanged.connect(self.slider_tick_changed)
        if checkable:
            self.slot_state_changed(0)

    def value_to_tick(self, value:float):
        return int((value - self.spinbox.minimum()) / (self.spinbox.maximum() - self.spinbox.minimum()) * (self.slider.maximum() - self.slider.minimum())) + self.slider.minimum()

    def tick_to_value(self, tick:int):
        return (tick - self.slider.minimum()) / (self.slider.maximum() - self.slider.minimum()) * (self.spinbox.maximum() - self.spinbox.minimum()) + self.spinbox.minimum()

    def spinbox_value_changed(self, value:float):
        if not self.should_call_slot:
            return
        
        old_should_call_slot = self.should_call_slot
        self.should_call_slot = False

        tick = self.value_to_tick(value)
        self.slider.setValue(tick)

        self.should_call_slot = old_should_call_slot

        self.value_changed.emit(value)

    def slider_tick_changed(self):
        if not self.should_call_slot:
            return
        
        old_should_call_slot = self.should_call_slot
        self.should_call_slot = False

        tick = self.slider.value()
        value = self.tick_to_value(tick)
        self.spinbox.setValue(value)

        self.should_call_slot = old_should_call_slot

        self.value_changed.emit(value)

    def slot_state_changed(self, state):
        self.slider.setEnabled(state == Qt.CheckState.Checked.value)
        self.spinbox.setEnabled(state == Qt.CheckState.Checked.value)
        self.checked_changed.emit(state == Qt.CheckState.Checked.value)

    @property
    def value(self):
        return self.spinbox.value()
    
    @value.setter
    def value(self, value:float):
        self.spinbox.setValue(value)

    @property
    def checked(self):
        if isinstance(self.label, QCheckBox):
            return self.label.isChecked()
        else:
            return True
    
    @checked.setter
    def checked(self, checked:bool):
        if isinstance(self.label, QCheckBox):
            self.label.setChecked(checked)


from PyQt6.QtWidgets import QWidget, QColorDialog, QHBoxLayout, QLabel
from PyQt6.QtGui import QColor, QMouseEvent, QPaintEvent, QPainter
from PyQt6.QtCore import pyqtSignal, Qt

import glm

class ColorChooser(QWidget):

    color_changed = pyqtSignal(QColor)
    glm_color_changed = pyqtSignal(object)

    def __init__(self, color:QColor=QColor(0,0,0), width:int=50, height:int=50, alpha:bool=False, parent:QWidget=None):
        QWidget.__init__(self, parent)
        self.__color = color
        self.__alpha = alpha
        self.setFixedSize(width, height)
    
    def mouseReleaseEvent(self, release_event: QMouseEvent | None) -> None:
        color = None
        if self.__alpha:
            color = QColorDialog.getColor(self.__color, self, "选择颜色", QColorDialog.ColorDialogOption.ShowAlphaChannel)
        else:
            color = QColorDialog.getColor(self.__color, self, "选择颜色")

        if color.isValid():
            self.color = color

        return super().mouseReleaseEvent(release_event)

    @property
    def color(self):
        return self.__color
    
    @color.setter
    def color(self, color:QColor):
        if self.__color != color:
            self.__color = color
            self.update()
            self.color_changed.emit(self.__color)
            self.glm_color_changed.emit(self.glm_color)
        
    @property
    def glm_color(self):
        if self.__alpha:
            return glm.vec4(self.__color.redF(), self.__color.greenF(), self.__color.blueF(), self.__color.alphaF())
        else:
            return glm.vec3(self.__color.redF(), self.__color.greenF(), self.__color.blueF())
        
    @glm_color.setter
    def glm_color(self, glm_color):
        if isinstance(glm_color, glm.vec3):
            self.color = QColor(int(glm_color.r*255), int(glm_color.g*255), int(glm_color.b*255))
        else:
            self.color = QColor(int(glm_color.r*255), int(glm_color.g*255), int(glm_color.b*255), int(glm_color.a*255))

    def paintEvent(self, paint_event: QPaintEvent | None) -> None:
        painter = QPainter(self)
        painter.setPen(Qt.PenStyle.NoPen)
        painter.setBrush(self.__color)
        painter.drawRect(self.rect())
        return super().paintEvent(paint_event)

class ColorInput(QWidget):

    color_changed = pyqtSignal(QColor)
    glm_color_changed = pyqtSignal(object)

    def __init__(self, prompt:str, color:QColor=QColor(0,0,0), width:int=50, height:int=50, alpha:bool=False, parent:QWidget=None):
        QWidget.__init__(self, parent=parent)

        hlayout = QHBoxLayout()
        hlayout.setAlignment(Qt.AlignmentFlag.AlignLeft)

        label = QLabel(prompt, self)
        hlayout.addWidget(label)

        self.__color_chooser = ColorChooser(color, width, height, alpha, parent)
        hlayout.addWidget(self.__color_chooser)
        self.setLayout(hlayout)

        self.__color_chooser.color_changed.connect(self.color_changed)
        self.__color_chooser.glm_color_changed.connect(self.glm_color_changed)

    @property
    def color(self)->QColor:
        return self.__color_chooser.color
    
    @color.setter
    def color(self, color:QColor)->None:
        self.__color_chooser.color = color

    @property
    def glm_color(self)->glm.vec3:
        return self.__color_chooser.glm_color
    
    @glm_color.setter
    def glm_color(self, glm_color:glm.vec3)->None:
        self.__color_chooser.glm_color = glm_color


import glm
import math
import uuid
import numpy as np

from glass.DictList import DictList
from glass.WeakSet import WeakSet
from glass.WeakDict import WeakDict
from glass.MetaInstancesRecorder import MetaInstancesRecorder
from .callback_vec import callback_quat, callback_vec3

class SceneNode(metaclass=MetaInstancesRecorder):

    @MetaInstancesRecorder.init
    def __init__(self, name:str="", unique_path:bool=False):
        if name:
            self._name = name
        else:
            self._name = self.__class__.__name__ + "_" + str(uuid.uuid1())

        self._unique_path:bool = unique_path
        self._position:callback_vec3 = callback_vec3(0, 0, 0, callback=self._set_dirty)
        self._orientation:callback_quat = callback_quat(1, 0, 0, 0, callback=self._update_yaw_pitch_roll)
        self._scale:callback_vec3 = callback_vec3(1, 1, 1, callback=self._set_dirty)
        self._yaw_pitch_roll:glm.vec3 = glm.vec3(0, 0, 0)

        self._parents:DictList = DictList(weak_ref=True)
        self._children:DictList = DictList()
        self._scenes:WeakSet = WeakSet()
        self._transform_dirty:WeakSet = WeakSet()
        self._children_transform_dirty:WeakDict = WeakDict(weak_ref_keys=True, weak_ref_values=False)
        self._should_update_yaw_pitch_roll:bool = True

        self._propagation_props:dict = {}
        self._block_propagation:set = set()

        self._propagation_props["visible"] = True

    @MetaInstancesRecorder.delete
    def __del__(self):
        pass

    @property
    def unique_path(self)->bool:
        return self._unique_path

    @unique_path.setter
    def unique_path(self, flag:bool)->None:
        self._unique_path = flag

    @property
    def visible(self):
        return self.propagation_prop("visible")
    
    @visible.setter
    def visible(self, visible:bool):
        self.set_propagation_prop("visible", visible)

    def hide(self):
        self.visible = False

    def show(self):
        self.visible = True

    @property
    def scenes(self):
        return self._scenes

    @property
    def parents(self):
        return self._parents

    @property
    def children(self):
        return self._children
    
    def block_propagation(self, name:str, flag:bool=True):
        if flag:
            self._block_propagation.add(name)
        elif name in self._block_propagation:
            self._block_propagation.remove(name)

    def propagation_prop(self, name:str):
        if name not in self._propagation_props:
            return None
        
        return self._propagation_props[name]

    def set_propagation_prop(self, name:str, value, callback=None, args:tuple=(), kwargs:dict={}):
        parent_queues = [self]
        is_self_set = False
        while parent_queues:
            parent = parent_queues.pop()
            if name in parent._block_propagation:
                child_queues = [parent]
                while child_queues:
                    child = child_queues.pop()
                    child._propagation_props[name] = value
                    if callback is not None:
                        callback(child, *args, **kwargs)
                    child_queues.extend(child._children)
                    if child is self:
                        is_self_set = True
            else:
                parent_queues.extend(parent.parents)

        if not is_self_set:
            child_queues = [self]
            while child_queues:
                child = child_queues.pop()
                child._propagation_props[name] = value
                if callback is not None:
                    callback(child, *args, **kwargs)
                child_queues.extend(child._children)

    def has_parent(self, parent):
        return (parent in self._parents)

    def has_child(self, child):
        return (child in self._children)

    def add_child(self, node):
        if node.has_parent(self):
            return

        if node.name in self._children:
            if self._children[node.name] is not node:
                old_name = node.name
                i = 1
                new_name = f"{old_name}-{i}"
                while new_name in self._children:
                    i += 1
                    new_name = f"{old_name}-{i}"
                node.name = new_name
            else:
                return

        if self.name in node._parents:
            if node._parents[self.name] is not self:
                raise NameError("already has one parent named '" + self.name + "'")
            else:
                return

        node._parents[self.name] = self
        self._children[node.name] = node
        node._add_as_child_callback()

        node._add_scenes(self._scenes)
        self._set_dirty(False, True)

    def remove_child(self, child):
        detached_node = None
        paths_prefix = None
        if isinstance(child, str):
            name = child
            if name in self._children:
                detached_node = self._children[name]
                paths_prefix = detached_node.paths_str
                del self._children[name]
                del detached_node._parents[self.name]
        else:
            try:
                self._children.remove(child)
                paths_prefix = child.paths_str
                child._parents.remove(self)
                detached_node = child
            except ValueError:
                pass

        if detached_node is None:
            return None
        
        detached_node._update_scenes()
        for scene in self.scenes:
            scene._remove_paths_prefix(paths_prefix)

    def clear_children(self):
        children_names = self.children_names
        for child_name in children_names:
            self.remove_child(child_name)

    @property
    def children_names(self):
        return list(self._children.keys())

    def __hash__(self):
        return id(self)

    def __getitem__(self, name:(str,int)):
        if isinstance(name, str):
            if name not in self._children:
                self.add_child(SceneNode(name))
        else:
            while name >= len(self._children):
                self.add_child(SceneNode())

        return self._children[name]

    def __setitem__(self, name:str, node):
        if name in self._children:
            if self._children[name] is node:
                return
            self.remove_child(name)

        node.name = name
        self.add_child(node)

    def __delitem__(self, name:str):
        self.remove_child(name)

    def __to_string(self, indent):
        result = indent * "  " + self.name
        for child in self._children:
            result += ("\n" + child.__to_string(indent+1))
        return result

    def __repr__(self):
        return self.__to_string(0)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name:str):
        if name == self._name:
            return

        if not self._parents:
            self._name = name
            return

        for parent in self._parents:
            if name in parent._children:
                raise NameError("parent node already has one child named '" + name + "'")

            del parent._children[self._name]
            parent._children[name] = self

        self._name = name

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, position:glm.vec3):
        self._position.x = position.x
        self._position.y = position.y
        self._position.z = position.z

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, scale:(glm.vec3, float)):
        if isinstance(scale, (int,float)):
            self._scale.x = scale
            self._scale.y = scale
            self._scale.z = scale
        else:
            self._scale.x = scale.x
            self._scale.y = scale.y
            self._scale.z = scale.z

    @property
    def orientation(self):
        return self._orientation

    @orientation.setter
    def orientation(self, orientation:glm.quat):
        self._orientation.w = orientation.w
        self._orientation.x = orientation.x
        self._orientation.y = orientation.y
        self._orientation.z = orientation.z

    def _add_scenes(self, scenes):
        self._scenes.update(scenes)
        for child in self._children:
            child._add_scenes(scenes)

    def _update_yaw_pitch_roll(self):
        if not self._should_update_yaw_pitch_roll:
            return
        
        q0 = self._orientation.w
        q1 = self._orientation.x
        q2 = self._orientation.y
        q3 = self._orientation.z
        self._yaw_pitch_roll[0] = math.atan2(2 * (q0*q3 - q1*q2), 1 - 2 * (q1*q1 + q3*q3))/math.pi*180
        self._yaw_pitch_roll[1] = math.asin(np.clip(2 * (q0*q1 + q2*q3), -1, 1))/math.pi*180
        self._yaw_pitch_roll[2] = math.atan2(2 * (q0*q2 - q1*q3), 1 - 2 * (q1*q1 + q2*q2))/math.pi*180

        self._set_dirty()

    def _update_orientation(self):
        yaw = self._yaw_pitch_roll[0]/180*math.pi
        pitch = self._yaw_pitch_roll[1]/180*math.pi
        roll = self._yaw_pitch_roll[2]/180*math.pi

        quat1 = glm.quat(math.cos(yaw / 2), 0, 0, math.sin(yaw / 2))
        quat2 = glm.quat(math.cos(pitch / 2), math.sin(pitch / 2), 0, 0)
        quat3 = glm.quat(math.cos(roll / 2), 0, math.sin(roll / 2), 0)

        self._should_update_yaw_pitch_roll = False
        self.orientation = quat1 * quat2 * quat3
        self._should_update_yaw_pitch_roll = True

        self._set_dirty()

    @property
    def yaw(self):
        return self._yaw_pitch_roll[0]

    @yaw.setter
    def yaw(self, yaw:float):
        self._yaw_pitch_roll[0] = yaw
        self._update_orientation()

    @property
    def pitch(self):
        return self._yaw_pitch_roll[1]

    @pitch.setter
    def pitch(self, pitch:float):
        self._yaw_pitch_roll[1] = pitch
        self._update_orientation()

    @property
    def roll(self):
        return self._yaw_pitch_roll[2]

    @roll.setter
    def roll(self, roll:float):
        self._yaw_pitch_roll[2] = roll
        self._update_orientation()
    
    @property
    def paths(self):
        if not self.parents:
            return [[self]]
        
        all_paths = []
        for parent in self.parents:
            parent_paths = parent.paths
            for path in parent_paths:
                path.append(self)
            all_paths.extend(parent_paths)

        return all_paths
    
    @property
    def paths_str(self):
        if not self.parents:
            return ["/" + self.name]
        
        all_paths_str = []
        for parent in self.parents:
            parent_paths_str = parent.paths_str
            for i, path_str in enumerate(parent_paths_str):
                parent_paths_str[i] = path_str + "/" + self.name
            all_paths_str.extend(parent_paths_str)

        return all_paths_str

    def _clear_dirty_scenes(self, scenes):
        for scene in scenes:
            if scene in self._transform_dirty:
                self._transform_dirty.remove(scene)

            if scene in self._children_transform_dirty:
                del self._children_transform_dirty[scene]

        for child in self._children:
            child._clear_dirty_scenes(scenes)

    def _update_scenes(self):
        old_scenes = self._scenes
        self._scenes = WeakSet()
        for parent in self.parents:
            self._scenes.update(parent._scenes)

        self._clear_dirty_scenes(old_scenes - self._scenes)

    def _set_upstream_dirty(self, scenes):
        for parent in self.parents:
            for scene in scenes:
                if scene not in parent._children_transform_dirty:
                    parent._children_transform_dirty[scene] = set()

                parent._children_transform_dirty[scene].add(self)

            parent._set_upstream_dirty(scenes)

    def _set_transform_dirty(self, scenes):
        return False

    def _set_dirty(self, self_transform_dirty=True, upstream=True, scenes=None):
        if scenes is None:
            scenes = self._scenes

        self_transform_dirty_set = False
        if self_transform_dirty:
            self_transform_dirty_set = self._set_transform_dirty(scenes)

        for scene in scenes:
            if self._children:
                if scene not in self._children_transform_dirty:
                    self._children_transform_dirty[scene] = set()

                for child in self._children:
                    self._children_transform_dirty[scene].add(child)
                    child._set_dirty(True, False, scenes)
            elif scene in self._children_transform_dirty:
                del self._children_transform_dirty[scene]

        if upstream and (self_transform_dirty_set or self._children):
            self._set_upstream_dirty(scenes)

    @property
    def parent(self):
        for parent in self._parents:
            return parent
    
    @property
    def scene(self):        
        for scene in self._scenes:
            return scene
        
    @property
    def abs_scale(self)->glm.vec3:
        return SceneNode.__abs_scale(self)
    
    @property
    def abs_orientation(self)->glm.quat:
        return SceneNode.__abs_orientation(self)
    
    @property
    def abs_position(self)->glm.vec3:
        return SceneNode.__abs_position(self)

    @property
    def path(self)->list:
        if self.parent is None:
            return [self]
        
        path = self.parent.paths[0]
        path.append(self)
        return path

    @staticmethod
    def __abs_orientation(node)->glm.quat:
        try:
            parent = node.parents[0]
        except IndexError:
            return node._orientation.flat

        parent_abs_orientation = SceneNode.__abs_orientation(parent)
        return parent_abs_orientation * node._orientation.flat

    @staticmethod
    def __abs_position(node)->glm.vec3:
        try:
            parent = node.parents[0]
        except:
            return node._position.flat

        parent_abs_orientation = SceneNode.__abs_orientation(parent)
        parent_abs_scale = SceneNode.__abs_scale(parent)
        parent_abs_position = SceneNode.__abs_position(parent)
        
        return parent_abs_orientation * (parent_abs_scale * node._position.flat) + parent_abs_position

    @staticmethod
    def __abs_scale(node)->glm.vec3:
        try:
            parent = node.parents[0]
        except:
            return node._scale.flat
        
        parent_abs_scale = SceneNode.__abs_scale(parent)
        return parent_abs_scale * node._scale.flat

    def _add_as_child_callback(self)->None:
        if self._unique_path:
            len_parents = len(self._parents)
            i = 0
            for parent in self._parents:
                if i >= len_parents-1:
                    break

                parent.remove_child(self)
                i += 1

            node = self
            while True:
                if len(node._parents) > 1:
                    raise RuntimeError('SceneNode can only have one path to root')
                
                try:
                    node = node._parents[0]
                except IndexError:
                    break

        for child in self._children:
            child._add_as_child_callback()


from PyQt6.QtWidgets import QWidget, QHBoxLayout, QLabel, QComboBox
from PyQt6.QtCore import pyqtSignal, Qt

from .SliderInput import SliderInput
from .ImageChooser import ImageChooser

from enum import Enum

class ValueImageInput(QWidget):

    class Type(Enum):
        Value = 0
        Image = 1

    type_changed = pyqtSignal(Type)
    value_changed = pyqtSignal(float)
    file_path_changed = pyqtSignal(str)

    def __init__(self, prompt:str="", parent:QWidget=None):
        QWidget.__init__(self, parent=parent)

        hlayout = QHBoxLayout()
        hlayout.setContentsMargins(0, 0, 0, 0)
        hlayout.setAlignment(Qt.AlignmentFlag.AlignLeft)

        label = QLabel(prompt, self)
        hlayout.addWidget(label)

        self.__type_combobox = QComboBox()
        self.__type_combobox.addItem("值", ValueImageInput.Type.Value)
        self.__type_combobox.addItem("贴图", ValueImageInput.Type.Image)
        hlayout.addWidget(self.__type_combobox)

        self.__slider_input = SliderInput(prompt="", range=(0, 1), value=0.5, parent=self)
        hlayout.addWidget(self.__slider_input)
        
        self.__image_chooser = ImageChooser(width=35, height=35, parent=self)
        hlayout.addWidget(self.__image_chooser)
        self.__image_chooser.hide()

        self.setLayout(hlayout)

        self.__type_combobox.currentIndexChanged.connect(self.__slot_type_changed)
        self.__slider_input.value_changed.connect(self.value_changed)
        self.__image_chooser.file_path_changed.connect(self.file_path_changed)

    def __slot_type_changed(self, index:int):
        if index == 0:
            self.__image_chooser.hide()
            self.__slider_input.show()
        else:
            self.__slider_input.hide()
            self.__image_chooser.show()
        self.type_changed.emit(self.__type_combobox.itemData(self.__type_combobox.currentIndex()))

    @property
    def type(self):
        return self.__type_combobox.itemData(self.__type_combobox.currentIndex())
    
    @type.setter
    def type(self, type:Type):
        if type == ValueImageInput.Type.Value:
            self.__type_combobox.setCurrentIndex(0)
        else:
            self.__type_combobox.setCurrentIndex(1)

    @property
    def value(self):
        return self.__slider_input.value
    
    @value.setter
    def value(self, value:float):
        self.__slider_input.value = value

    @property
    def file_path(self):
        return self.__image_chooser.file_path

    @file_path.setter
    def file_path(self, file_path:str):
        self.__image_chooser.file_path = file_path


import os
import shutil

class GlassConfig:

    def __init__(self):
        self_folder = os.path.dirname(os.path.abspath(__file__)).replace("\\", "/")
        glass_engine_glsl_folder = self_folder + "/../glass_engine/glsl"
        glass_glsl_folder = self_folder + "/glsl"
        if os.path.isdir(glass_engine_glsl_folder):
            self.__cache_folder = glass_engine_glsl_folder + "/__glcache__"
        else:
            self.__cache_folder = glass_glsl_folder + "/__glcache__"

        self.debug:bool = False
        self.print:bool = True
        self.warning:bool = True
        self.recompile:bool = False

    @property
    def cache_folder(self)->str:
        if not os.path.isdir(self.__cache_folder):
            os.makedirs(self.__cache_folder)

        return self.__cache_folder

    @cache_folder.setter
    def cache_folder(self, folder_path:str):
        self.__cache_folder = folder_path

    def clear_cache(self):
        if os.path.isdir(self.__cache_folder):
            shutil.rmtree(self.__cache_folder)

GlassConfig = GlassConfig()


class CSRMat:

    def __init__(self, rows:int, cols:int=None, eye:bool=False)->None:
        if cols is None:
            cols = rows

        self._cols = cols
        if eye:
            n = min(rows, cols)

            self.indptr = list(range(n+1))
            if rows > n:
                self.indptr.extend((rows+1-n)*[n])

            self.col_indices = list(range(n))
            self.row_indices = list(range(n))
            self.data = n*[1]
        else:
            self.indptr = (rows+1)*[0]
            self.col_indices = []
            self.row_indices = []
            self.data = []

    def __getitem__(self, ij)->object:
        i, j = ij
        
        start_index = self.indptr[i]
        end_index = self.indptr[i+1]
        cols = self.col_indices[start_index:end_index]
        try:
            index_j = cols.index(j)
            return self.data[start_index+index_j]
        except:
            return 0
        
    def __setitem__(self, ij:tuple, value:object)->None:
        if value == 0:
            self.__delitem__(ij)
            return

        i, j = ij
        
        start_index = self.indptr[i]
        end_index = self.indptr[i+1]
        cols = self.col_indices[start_index:end_index]
        try:
            index_j = cols.index(j)
            self.data[start_index+index_j] = value
        except:
            self.col_indices.insert(self.indptr[i], j)
            self.row_indices.insert(self.indptr[i], i)
            self.data.insert(self.indptr[i], value)
            for k in range(i+1, len(self.indptr)):
                self.indptr[k] += 1

    def __delitem__(self, ij:tuple)->None:
        i, j = ij

        if i < 0 or i >= self.rows:
            raise IndexError(j)
        
        if j < 0 or j >= self.cols:
            raise IndexError(j)

        start_index = self.indptr[i]
        end_index = self.indptr[i+1]
        cols = self.col_indices[start_index:end_index]
        try:
            index_j = cols.index(j)
            del self.data[start_index+index_j]
            del self.col_indices[start_index+index_j]
            del self.row_indices[start_index+index_j]
            for k in range(i+1, len(self.indptr)):
                self.indptr[k] -= 1
        except:
            pass

    @property
    def rows(self)->int:
        return self.indptr.__len__() - 1
    
    @property
    def cols(self)->int:
        return self._cols
    
    def __repr__(self)->str:
        result = ''
        rows = self.rows
        cols = self.cols
        for i in range(rows):
            for j in range(cols):
                result += str(self[i,j])
                if j < cols-1:
                    result += ' '
            if i < rows-1:
                result += "\n"
        return result
    
    def insert_row(self, i:int, n:int=1)->None:
        for _ in range(n):
            self.indptr.insert(i, self.indptr[i])

    def append_row(self, n:int=1)->None:
        for _ in range(n):
            self.indptr.insert(self.rows, self.indptr[self.rows])

    def reset_row(self, row_start:int, row_end:int=None)->None:
        if row_end is None:
            row_end = row_start + 1

        if row_end == row_start:
            return
        
        start = self.indptr[row_start]
        stop = self.indptr[row_end]
        del self.col_indices[start:stop]
        del self.row_indices[start:stop]
        del self.data[start:stop]
        delta = stop - start
        for it in range(row_end, len(self.indptr)):
            self.indptr[it] -= delta

    def remove_row(self, row_start:int, row_end:int=None)->None:
        if row_end is None:
            row_end = row_start + 1

        if row_end == row_start:
            return
        
        del self.col_indices[self.indptr[row_start]:self.indptr[row_end]]
        del self.row_indices[self.indptr[row_start]:self.indptr[row_end]]
        del self.data[self.indptr[row_start]:self.indptr[row_end]]
        del self.indprt[row_start:row_end]



import glm

class callback_vec3(glm.vec3):

    def __init__(self, x:float=0.0, y:float=0.0, z:float=0.0, callback=None):
        glm.vec3.__init__(self, x, y, z)
        self.__dict__["_callback"] = callback
    
    def __deepcopy__(self, memo):
        return self.flat
    
    def __copy__(self):
        return self.flat

    @property
    def flat(self):
        return glm.vec3(self.x, self.y, self.z)

    def __setattr__(self, name:str, value):
        super().__setattr__(name, value)
        if name in "xyzrgbstp" and self._callback is not None:
            self._callback()

    def __rmul__(self, other):
        return other * self.flat
    
class callback_vec4(glm.vec4):

    def __init__(self, x:float=0.0, y:float=0.0, z:float=0.0, w:float=1.0, callback=None):
        glm.vec4.__init__(self, x, y, z, w)
        self.__dict__["_callback"] = callback
    
    def __deepcopy__(self, memo):
        return self.flat

    def __copy__(self):
        return self.flat

    @property
    def flat(self):
        return glm.vec4(self.x, self.y, self.z, self.w)

    def __setattr__(self, name:str, value):
        super().__setattr__(name, value)
        if name in "xyzwrgbastpq" and self._callback is not None:
            self._callback()

    def __rmul__(self, other):
        return other * self.flat

class callback_quat(glm.quat):
    def __init__(self, w:float=1.0, x:float=0.0, y:float=0.0, z:float=0.0, callback=None):
        glm.quat.__init__(self, w, x, y, z)
        self.__dict__["_callback"] = callback
    
    @property
    def flat(self):
        return glm.quat(self.w, self.x, self.y, self.z)

    def __deepcopy__(self, memo):
        return self.flat
    
    def __copy__(self):
        return self.flat

    def __setattr__(self, name:str, value):
        super().__setattr__(name, value)
        if name in "wxyz" and self._callback is not None:               
            self._callback()

    def __mul__(self, other):
        return self.flat * other
    
    def __rmul__(self, other):
        return other * self.flat
    


from OpenGL import GL
import OpenGL.GL.ARB.bindless_texture as bt

from .FBOAttachment import FBOAttachment
from .GLInfo import GLInfo
from .utils import checktype
from .helper import get_dtype

class sampler2DMS(FBOAttachment):

	_basic_info = \
	{
		"gen_func": GL.glGenTextures,
		"bind_func": GL.glBindTexture,
		"del_func": GL.glDeleteTextures,
		"target_type": GL.GL_TEXTURE_2D_MULTISAMPLE,
		"binding_type": GL.GL_TEXTURE_BINDING_2D_MULTISAMPLE,
		"need_number": True,
	}

	def __init__(self, width:int=0, height:int=0, samples:int=4, internal_format:GLInfo.internal_formats=GL.GL_RGBA32F):
		FBOAttachment.__init__(self)
		
		self._handle = 0
		self._samples = samples
		self._width = width
		self._height = height
		self._internal_format = internal_format
		self._param_changed = True

	def __del__(self):
		if self._handle != 0:
			self._handle = 0

		FBOAttachment.__del__(self)

	def __deepcopy__(self, memo):
		result = sampler2DMS()
		result._samples = self._samples
		result._width = self._width
		result._height = self._height
		result._internal_format = self._internal_format

		return result

	def bind(self, update_fbo:bool=False, force_update_image:bool=False):
		FBOAttachment.bind(self, update_fbo, force_update_image)
		if force_update_image or self._param_changed:
			GL.glTexImage2DMultisample(GL.GL_TEXTURE_2D_MULTISAMPLE, self._samples, self._internal_format, self._width, self._height, GL.GL_TRUE)
			self._param_changed = False

	@property
	def handle(self):
		if not bt.glGetTextureHandleARB:
			return 0

		if self._id == 0:
			return 0
		
		if self._handle == 0:
			self._handle = bt.glGetTextureHandleARB(self._id)
			if self._handle == 0:
				raise RuntimeError(f"failed to create sampler2D {self._id}'s handle")
			self._dynamic = False
			bt.glMakeTextureHandleResidentARB(self._handle)
			
		return self._handle

	@property
	def samples(self):
		return self._samples

	@samples.setter
	@FBOAttachment.param_setter
	def samples(self, samples:int):
		self._samples = samples
		self._param_changed = True

	@property
	def width(self):
		return self._width

	@width.setter
	@FBOAttachment.param_setter
	def width(self, width:int):
		self._width = width
		self._param_changed = True

	@property
	def height(self):
		return self._height

	@height.setter
	@FBOAttachment.param_setter
	def height(self, height:int):
		self._height = height
		self._param_changed = True

	@property
	def dtype(self):
		return get_dtype(self._internal_format)

	@property
	def internal_format(self):
		return self._internal_format

	@internal_format.setter
	@FBOAttachment.param_setter
	def internal_format(self, format:GLInfo.internal_formats):
		self._internal_format = format
		self._param_changed = True

	def update(self):
		self._param_changed = True

	@checktype
	def malloc(self, width:int, height:int, samples:int=4, layers:int=None, internal_format:GLInfo.internal_formats=None):
		self.width = width
		self.height = height
		self.samples = samples

		if internal_format is None:
			internal_format = self.internal_format

		if internal_format is None:
			internal_format = GL.GL_RGBA32F

		self.internal_format = internal_format

	@property
	def is_completed(self):
		return (self._width > 0 and self._height > 0)



from glass.download import pip_install
from .QtScreen import init_QtScreen
from ..Manipulators.Manipulator import Manipulator

try:
    import PyQt6
    from PyQt6.QtOpenGLWidgets import QOpenGLWidget
except ModuleNotFoundError:
    pip_install("PyQt6")
    import PyQt6
    from PyQt6.QtOpenGLWidgets import QOpenGLWidget

import glm

@init_QtScreen
class PyQt6Screen(QOpenGLWidget):
    mouse_pressed = PyQt6.QtCore.pyqtSignal(Manipulator.MouseButton, glm.vec2, glm.vec2)
    mouse_released = PyQt6.QtCore.pyqtSignal(Manipulator.MouseButton, glm.vec2, glm.vec2)
    mouse_double_clicked = PyQt6.QtCore.pyqtSignal(Manipulator.MouseButton, glm.vec2, glm.vec2)
    mouse_moved = PyQt6.QtCore.pyqtSignal(glm.vec2, glm.vec2)
    wheel_scrolled = PyQt6.QtCore.pyqtSignal(glm.vec2, glm.vec2, glm.vec2)
    key_pressed = PyQt6.QtCore.pyqtSignal(Manipulator.Key)
    key_released = PyQt6.QtCore.pyqtSignal(Manipulator.Key)
    key_repeated = PyQt6.QtCore.pyqtSignal(set)
    frame_started = PyQt6.QtCore.pyqtSignal()
    frame_ended = PyQt6.QtCore.pyqtSignal()



from .sampler2DMS import sampler2DMS
from .GLInfo import GLInfo
from .utils import checktype

from OpenGL import GL
import OpenGL.GL.ARB.bindless_texture as bt

class usampler2DMS(sampler2DMS):

    @checktype
    def __init__(self, width:int=0, height:int=0, samples:int=4, internal_format:GLInfo.usampler_internal_formats=None):
        if internal_format is None:
            internal_format = GL.GL_RGBA32UI

        sampler2DMS.__init__(self, width, height, samples, internal_format)

    @property
    def internal_format(self):
        return self._internal_format

    @internal_format.setter
    @checktype
    def internal_format(self, format:GLInfo.usampler_internal_formats):
        if self._internal_format != format:
            self._internal_format = format
            self._param_changed = True

    @property
    def handle(self):
        if not bt.glGetTextureHandleARB:
            return 0
        
        self.bind()
        if self._handle == 0:
            self._handle = bt.glGetTextureHandleARB(self._id)
            if self._handle == 0:
                raise RuntimeError(f"failed to create isampler2D {self._id}'s handle")
            bt.glMakeTextureHandleResidentARB(self._handle)
            self._dynamic = False

        return self._handle


from .Light import Light, FlatLight
from ..algorithm import fzero

from glass.utils import checktype
from glass.DictList import DictList
from glass.ShaderStorageBlock import ShaderStorageBlock

import glm
import math

class PointLight(Light):

    @checktype
    def __init__(self, name:str=""):
        Light.__init__(self, name)
        
        self._K1 = 0.045
        self._K2 = 0.0075

        self._coverage = 0
        self.__update_coverage()

    def __update_coverage(self):
        epsilon = 0.01
        inv_e = 1/epsilon
        self._coverage = (-self._K1 + math.sqrt(self._K1**2 + 4*self._K2*(inv_e-1)))/(2*self._K2)
    
        for flat in self._flats:
            flat.coverage = self._coverage

        self._update_scene_lights()

    @property
    def coverage(self):
        return 0.1*self._coverage
    
    @coverage.setter
    def coverage(self, coverage:float):
        def func(t):
            K1 = 3.651720188286232 / math.pow(t - 1.379181323137789, 0.956790970458513)
            K2 = 27.101525310782399 / math.pow(t - 2.191989674193149, 1.727016118197271)
            epsilon = 0.01
            inv_e = 1/epsilon
            return (-K1 + math.sqrt(K1**2 + 4*K2*(inv_e-1)))/(2*K2) - 10*coverage
        
        t = fzero(func, [2.2, float("inf")])
        if t is None:
            t = 2.2

        self._K1 = 3.651720188286232 / math.pow(t - 1.379181323137789, 0.956790970458513)
        self._K2 = 27.101525310782399 / math.pow(t - 2.191989674193149, 1.727016118197271)
        for flat in self._flats:
            flat.K1 = self._K1
            flat.K2 = self._K2

        self.__update_coverage()

    @property
    def K1(self):
        return self._K1
    
    @K1.setter
    @checktype
    def K1(self, K1:float):
        if self._K1 == K1:
            return
        
        self._K1 = K1
        for flat in self._flats:
            flat.K1 = self._K1
            
        self.__update_coverage()

    @property
    def K2(self):
        return self._K2
    
    @K2.setter
    @checktype
    def K2(self, K2:float):
        if self._K2 == K2:
            return
        
        self._K2 = K2
        for flat in self._flats:
            flat.K2 = self._K2

        self.__update_coverage()

class FlatPointLight(FlatLight):

    def __init__(self, point_light:PointLight):
        self.abs_position = glm.vec3(0, 0, 0)
        self.need_update_depth_map = True
        FlatLight.__init__(self, point_light)

    def update(self, point_light:PointLight):
        self.K1 = point_light._K1
        self.K2 = point_light._K2
        self.coverage = point_light._coverage
        FlatLight.update(self, point_light)

class PointLights(ShaderStorageBlock.HostClass):

    def __init__(self):
        ShaderStorageBlock.HostClass.__init__(self)
        self.point_lights = DictList()

    def __getitem__(self, path_str:str):
        return self.point_lights[path_str]
    
    @ShaderStorageBlock.HostClass.not_const
    def __setitem__(self, path_str:str, point_light:FlatPointLight):
        self.point_lights[path_str] = point_light

    @ShaderStorageBlock.HostClass.not_const
    def __delitem__(self, path_str:str):
        del self.point_lights[path_str]

    def __contains__(self, path_str:str):
        return (path_str in self.point_lights)
    
    def __len__(self):
        return len(self.point_lights)
    
    def __iter__(self):
        return iter(self.point_lights)
    
    def keys(self):
        return self.point_lights.keys()

    @property
    def n_point_lights(self):
        return len(self.point_lights)
    
    